#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu March 1 2017

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from .... import pyqtgraphCore as pg
import numpy as np
import pandas as pd
from .template import *
from ....analysis.data_types import Transmission
from ....analysis.history_widget import HistoryTreeWidget
import traceback
from functools import partial
from ....common import get_window_manager


class PeaksItemGraph(pg.GraphItem):
    """
    Scatter plot to display the peaks and bases as dots which are editable.
    """
    def __init__(self):
        pg.GraphItem.__init__(self)
        self.dragPoint = None
        self.dragOffset = None
        self.scatter.sigClicked.connect(self.clicked)
        self.orig_data = None
        self.mode = 'view'
        self._brush_size = 15
        
        self.changed = False
    
    def set_mode(self, mode: str):
        self.mode = mode
        
    def setData(self, **kwds):
        self.text = kwds.pop('text', [])
        if 'curve_data' in kwds.keys():
            self.curve_data = kwds.pop('curve_data')
        self.data = kwds

        if 'events' in kwds.keys():
            self.events = kwds.pop('events')
        
        if 'pos' in self.data:
            npts = self.data['pos'].shape[0]
            self.data['data'] = np.empty(npts, dtype=[('index', int)])
            self.data['data']['index'] = np.arange(npts)
#        self.setTexts(self.text)
        self.updateGraph()


    def setTexts(self, text):
        for i in self.textItems:
            i.scene().removeItem(i)
        self.textItems = []
        for t in text:
            item = pg.TextItem(t)
            self.textItems.append(item)
            item.setParentItem(self)
        
    def updateGraph(self):
        pg.GraphItem.setData(self, **self.data)
        if hasattr(self, '_brush_size'):
            self.scatter.setSize(self._brush_size)
#        for i,item in enumerate(self.textItems):
#            item.setPos(*self.data['pos'][i])
    
    def drag(self, ev):
        if ev.button() != QtCore.Qt.LeftButton:
            ev.ignore()
            return
        
        if ev.isStart():
            # We are already one step into the drag.
            # Find the point(s) at the mouse cursor when the button was first 
            # pressed:
            pos = ev.buttonDownPos()
            pts = self.scatter.pointsAt(pos)
            if len(pts) == 0:
                ev.ignore()
                return
            self.dragPoint = pts[0]
            ind = pts[0].data()[0]
            self.dragOffset = self.data['pos'][ind] - pos
            
#            print(self.data)
#            print(self.dragPoint.data()[0])
            
        elif ev.isFinish():
            self.dragPoint = None
            return
        else:
            if self.dragPoint is None:
                ev.ignore()
                return
        
        ind = self.dragPoint.data()[0]
        self.data['pos'][ind] = ev.pos() + self.dragOffset

        x_pos = int(self.data['pos'][ind][0])
        
        if x_pos < 0:
            x_pos = 1
            self.data['pos'][ind][0] = float(x_pos)

        elif x_pos > self.curve_data.shape[0]:
            x_pos = self.curve_data.shape[0] - 1
            self.data['pos'][ind][0] = float(x_pos)

        self.data['pos'][ind][1] = self.curve_data[x_pos]
        self.updateGraph()
        ev.accept()
        
        self.changed = True
    
    def get_edited_dataframe(self) -> pd.DataFrame:
        d = {'event': self.data['pos'][:, 0].astype(np.int64), 'label': self.events}
        df = pd.DataFrame(d).sort_values(by=['event']).reset_index(drop=True)
        return df
    
    def delete_item(self, pt):
        ix = pt.data()[0]
        data = self.data.copy()
        data['pos'] = np.delete(self.data['pos'], ix, axis=0)
        del data['symbolBrush'][ix]
        del self.events[ix]
#        print(self.data)
#        self.updateGraph()
        self.setData(**data)
        self.changed = True

    def delete_all_items(self):
        # while self.scatter.points().size > 1:
        #     self.delete_item(self.scatter.points()[0])
        self.scatter.clear()
        for ix in range(len(self.events)):
            del self.events[0]
    
    def add_item(self, pos):
        x_pos = pos[0]
        y_pos = self.curve_data[x_pos]
        
        item_pos = np.array([x_pos, y_pos])
        
        data = self.data.copy()
        data['pos'] = np.vstack((data['pos'], item_pos))
        
        if self.mode == 'add_peaks':
            event = 'peak'
            brush = pg.fn.mkBrush(255, 0, 0, 150)
            
        elif self.mode == 'add_bases':
            event = 'base'
            brush = pg.fn.mkBrush(0, 255, 0, 150)
        
        self.events.append(event)
        data['symbolBrush'] += [brush]
        
        self.setData(**data)
        self.changed = True
    
    def mouseDragEvent(self, ev):
        if self.mode == 'drag':
            self.drag(ev)
        else:
            ev.ignore()
            return
#        elif self.mode == 'add':
#            print(ev.buttonDownPos())
#        else:
#            ev.ignore()
#    def mouseClickEvent(self, ev):
#        
    
    def clicked(self, scatter_plot, spot_items):
#        print("clicked: %s" % spot_items)
        
        if self.mode == 'delete':
#            print(spot_items)
            if not len(spot_items) == 1:
                return
            self.delete_item(spot_items[0])
    
    def set_brush_size(self, size: int):
        self._brush_size = size
        self.scatter.setSize(size)
            

class PeakEditorWindow(QtWidgets.QMainWindow, Ui_MainWindow):
    sig_disconnect_flowchart = QtCore.pyqtSignal()
    sig_reconnect_flowchart = QtCore.pyqtSignal()
    sig_send_data = QtCore.pyqtSignal(Transmission)
    
    def __init__(self, trans_curves, trans_peaks_bases):
        # super().__init__()
        super(PeakEditorWindow, self).__init__()
        # Ui_MainWindow.__init__(self)
        self.setupUi(self)
        self.tpb = None

        self.connected = True
        self.current_curve = None
        self.btnDisconnectFromFlowchart.setStyleSheet("background-color: yellow")
#        self.current_scatter_plots = None
        self.current_peak_scatter_plot = None
        
        self.mode_buttons = {'view': self.btnModeView, 
                             'drag': self.btnModeDrag,
                             'add_peaks': self.btnModeAddPeaks,
                             'add_bases': self.btnModeAddBases,
                             'delete': self.btnModeDelete
                             }
        
        for btn in self.mode_buttons.values():
            btn.setStyleSheet("background-color: red")
        
        self.current_mode = 'view'
        self.mode_buttons['view'].setStyleSheet("background-color: green")
        self.btnModeView.setChecked(True)
        
        self.connect_mode_buttons()
        
        self.brush_size = 15

        self.history_tree = HistoryTreeWidget()
        self.history_tree.fill_widget(trans_peaks_bases.history_trace.get_all_data_blocks_history)
        self.update_transmission(trans_curves, trans_peaks_bases)
        self._reset_listw()
        self.listwIndices.currentItemChanged.connect(self._set_row)
        self.listwIndices.setCurrentRow(0)
#        self.listwIndices.itemClicked.connect(self._set_row)
        self.sliderDotSize.valueChanged.connect(self._set_pens)
        self.dockWidget.setWidget(self.history_tree)
        
        self.graphicsView.sigMouseClicked.connect(lambda gv, coors: self._add_event(self.current_curve.getViewBox().mapSceneToView(coors).toPoint()))
        
        self.btnSaveCurve.clicked.connect(self.save_current_curve)
        # self.btnClearCurve.clicked.connect(self.clear_current_curve)

        self.btnDisconnectFromFlowchart.clicked.connect(self._disconnect_flowchart)
        self.btnSendToFlowchart.clicked.connect(self._send_data_to_flowchart)

        self.pushButtonOpenInViewer.clicked.connect(self.open_current_curve_in_viewer)

    def _disconnect_flowchart(self):
        self.connected = False
        self.btnDisconnectFromFlowchart.setText('Re-connect to flowchart')
        self.btnDisconnectFromFlowchart.setStyleSheet("background-color: orange")
        self.btnDisconnectFromFlowchart.clicked.disconnect(self._disconnect_flowchart)
        self.btnDisconnectFromFlowchart.clicked.connect(self._reconnect_flowchart)
        self.sig_disconnect_flowchart.emit()
    
    def _reconnect_flowchart(self):
        if QtWidgets.QMessageBox.question(self, 'RECONNECT?', 'Are you sure you want to reconnect to the flowchart?\n'
                                                              'YOU WILL LOOSE ALL MODFICATIONS THAT YOU HAVE MADE!') == QtWidgets.QMessageBox.No:
            return
        # self.btnDisconnectFromFlowchart.setChecked(False)
        self.btnDisconnectFromFlowchart.clicked.disconnect(self._reconnect_flowchart)
        self.btnDisconnectFromFlowchart.clicked.connect(self._disconnect_flowchart)
        self.btnDisconnectFromFlowchart.setText('Disconnect from flowchart')
        self.btnDisconnectFromFlowchart.setStyleSheet("background-color: yellow")
        self.connected = True
        self.sig_reconnect_flowchart.emit()
    
    def _send_data_to_flowchart(self):
        if self.get_data() is None:
            QtWidgets.QMessageBox.warning(self, 'Nothing to send', 'No data has been loaded that can be transmitted')
            return

        self.save_current_curve()

        to_send = self.get_data().copy()
        to_send.df = to_send.df[~to_send.df.apply(lambda r: r.peaks_bases.empty, axis=1)]
        self.sig_send_data.emit(to_send)
        
    def connect_mode_buttons(self):
        for k in self.mode_buttons.keys():
            self.mode_buttons[k].clicked.connect(partial(self.set_mode, k))
        
    def set_mode(self, mode):
        self.mode_buttons[self.current_mode].setStyleSheet("background-color: red")
        self.mode_buttons[self.current_mode].setChecked(False)
        
        self.mode_buttons[mode].setStyleSheet("background-color: green")
        self.current_mode = mode
        
        self.current_peak_scatter_plot.set_mode(mode)

    
    def _add_event(self, qpoint: QtCore.QPoint):
        if self.current_mode not in ['add_peaks', 'add_bases']:
            return
        self.current_peak_scatter_plot.add_item((qpoint.x(), qpoint.y()))

    def update_transmission(self, trans_curves: Transmission, trans_peaks_bases: Transmission):
        if not self.connected:
            return
        
        if hasattr(self, 'tc'):
            if self.tc.df.index.size != trans_curves.df.index.size:
                self.tc = trans_curves
                self._reset_listw()

            elif self.tc != trans_curves:
                self.tc = trans_curves
                self._set_row()
        else:
            self.tc = trans_curves

        self.tpb = trans_peaks_bases.copy()
        self._set_row()
        self.history_tree.fill_widget(trans_peaks_bases.history_trace.get_all_data_blocks_history())

    def _reset_listw(self):
        self.listwIndices.clear()
        self.listwIndices.addItems(list(map(str, [*range(self.tc.df.index.size)])))
    
    def save_current_curve(self):
        ix = self.current_ix
        self.tpb.df.at[ix, 'peaks_bases'] = self.current_peak_scatter_plot.get_edited_dataframe()
        self.current_peak_scatter_plot.changed = False

    def clear_current_curve(self):
        self.current_peak_scatter_plot.delete_all_items()
        self.save_current_curve()

    def _set_row(self):
        if self.current_peak_scatter_plot is not None:
            if self.current_peak_scatter_plot.changed:
                self.save_current_curve()

        self.graphicsView.clear()

        ix = self.listwIndices.currentRow()
        self.current_ix = ix
        curve_plot = pg.PlotDataItem()
        self.current_curve = curve_plot

        curve = self.tc.df['curve'].iloc[ix]

        sample_id = self.tc.df['SampleID'].iloc[ix]
        if not isinstance(sample_id, str):
            try:
                sample_id = str(sample_id)
            except:
                print("Could not display SampleID for curve at index: " + str(ix))
        self.lineEditSampleID.setText(sample_id)

        curve_plot.setData(curve)

        self.graphicsView.addItem(curve_plot)


        try:
            bases = self.tpb.df['peaks_bases'].iloc[ix]['event'][self.tpb.df['peaks_bases'].iloc[ix]['label'] == 'base'].values
            peaks = self.tpb.df['peaks_bases'].iloc[ix]['event'][self.tpb.df['peaks_bases'].iloc[ix]['label'] == 'peak'].values#tolist()

        except Exception:
            QtWidgets.QMessageBox.warning(self, 'Warning', 'The curve probably contains no peaks.\n' +
                                          traceback.format_exc())
            peaks = np.array([], dtype=int)
            bases = np.array([], dtype=int)

        xs = peaks; ys = np.take(curve, peaks)
        peaks_data = np.column_stack((xs, ys))

        xs_b = bases; ys_b = np.take(curve, bases)
        bases_data = np.column_stack((xs_b, ys_b))


        events = ((['base'] * bases_data.shape[0] + ['peak'] * peaks_data.shape[0]))

        all_data = np.vstack((bases_data, peaks_data))
        red_brushes = [pg.fn.mkBrush(255, 0, 0, 150)] * peaks_data.shape[0]
        green_brushes = [pg.fn.mkBrush(0, 255, 0, 150)] * bases_data.shape[0]
        all_brushes = green_brushes + red_brushes


        peaks_plot = PeaksItemGraph()
        peaks_plot.set_mode(self.current_mode)

        peaks_plot.setData(pos=all_data, size=self.brush_size, symbolBrush=all_brushes, curve_data=curve, events=events)# / min_curve)

        self.graphicsView.addItem(peaks_plot)
        self.current_peak_scatter_plot =peaks_plot

    def open_current_curve_in_viewer(self):
        w = get_window_manager().get_new_viewer_window()

        row = self.tc.df.iloc[self.current_ix]
        proj_path = self.tc.get_proj_path()
        w.open_from_dataframe(proj_path, row=row)

    def _set_pens(self, n):
        self.current_peak_scatter_plot.set_brush_size(n)
        self.brush_size = n


    def get_data(self) -> Transmission:
        return self.tpb



# -*- coding: utf-8 -*-
"""
Created on June 20 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import abc
import numpy as np
from PyQt5 import QtWidgets
from .... import pyqtgraphCore as pg
from ....common import get_proj_config, NoProjectOpen, InheritDocs
from warnings import warn
from copy import deepcopy
from typing import Union, List


class _AbstractBaseROI(metaclass=InheritDocs):
    """
    Abstract base class defining an ROI that works with the ROIList and ROI Managers.
    Inherit from this or BaseROI to make a new ROI class
    """
    @abc.abstractmethod
    def __init__(self, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, state: Union[dict, None]):
        """
        Minimum required attributes

        :param curve_plot_item: The plot item that is used for display the curves in the viewer
        :param view_box:        ViewBox containing the image sequence, used for overlaying the ROIs on top of the image
        :param state:           ROI state, used for restoring the ROIs. Pass None is not restoring an ROI from a state dict
        """
        pass

    # @property
    # @abc.abstractmethod
    # def curve_data(self) -> tuple:
    #     """
    #     The curve data for this ROI
    #
    #     :return: (x, y), [np.ndarray, np.ndarray]
    #     :rtype: tuple
    #     """
    #     pass
    #
    # @curve_data.setter
    # @abc.abstractmethod
    # def curve_data(self, data: tuple):
    #     """
    #     Set the curve data for this ROI
    #
    #     :param data: (x, y), [np.ndarray, np.ndarray]
    #     """
    #     pass

    @abc.abstractmethod
    def get_roi_graphics_object(self) -> QtWidgets.QGraphicsObject:
        """Get the QGraphicsObject used for visualization of the spatial localization of the ROI"""
        pass

    @abc.abstractmethod
    def set_roi_graphics_object(self, *args, **kwargs):
        """Set the QGraphicsObject used for visualization of the spatial localization of the ROI"""
        pass

    @abc.abstractmethod
    def reset_color(self):
        """Reset the color of this ROI back to the original color"""
        pass

    @abc.abstractmethod
    def set_original_color(self, color):
        """
        Set the original color for this ROI

        :param color: 1D numpy array of 4 floating point numbers (range 0 - 255) in RBGA format, [R, G, B, A]
        """
        pass

    @abc.abstractmethod
    def get_color(self) -> np.ndarray:
        """
        Get the current color of this ROI

        :return: 1D numpy array of 4 floating point numbers (range 0 - 255) in RBGA format, [R, G, B, A]
        :rtype: np.ndarray
        """
        pass

    @abc.abstractmethod
    def set_color(self, color, *args, **kwargs):
        """
        Set the current color of this ROI

        :param color: 1D numpy array of 4 floating point numbers (range 0 - 255) in RBGA format, [R, G, B, A]
        """
        pass

    @abc.abstractmethod
    def set_text(self, text: str):
        """Not implemented"""
        pass

    @abc.abstractmethod
    def set_tag(self, roi_def: str, tag: str):
        """
        Set a tag for the passed roi_def

        :param roi_def: The ROI_DEF that should be tagged
        :param tag:     The tag to label for the passed ROI_DEF/ROI Type
        """
        pass

    @abc.abstractmethod
    def get_tag(self, roi_def) -> str:
        """
        Get the tag that is set to the passed 'roi_def'

        :rtype: str
        """
        pass

    @abc.abstractmethod
    def get_all_tags(self) -> dict:
        """
        Get all the tags for all the ROI_DEFs

        :rtype: dict
        """
        pass

    @abc.abstractmethod
    def add_to_viewer(self):
        """Add this ROI to the viewer."""
        pass

    @abc.abstractmethod
    def remove_from_viewer(self):
        """Remove this ROI from the viewer"""
        pass

    @abc.abstractmethod
    def to_state(self):
        """Get the current state for this ROI so that it can be restored later"""
        pass

    @classmethod
    @abc.abstractmethod
    def from_state(cls, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, state: dict):
        """
        Restore this ROI from a state

        :param curve_plot_item: The plot item that is used for display the curves in the viewer
        :param view_box:        ViewBox containing the image sequence, used for overlaying the ROIs on top of the image
        :param state:           ROI state, used for restoring the ROIs. Pass None is not restoring an ROI from a state dict
        """
        pass


class BaseROI(_AbstractBaseROI):
    """
    A base class that is used by ManualROI and CNMFEROI
    Inherit from this to make a new ROI class
    """
    def __init__(self, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, state: Union[dict, None] = None,
                 metadata: dict = None):
        """
        Instantiate common attributes

        :param curve_plot_item: The plot item that is used for display the curves in the viewer
        :param view_box:        ViewBox containing the image sequence, used for overlaying the ROIs on top of the image
        :param state:           ROI state, used for restoring the ROIs. Pass None is not restoring an ROI from a state dict

        """
        super().__init__(curve_plot_item, view_box, state)
        if isinstance(curve_plot_item, pg.PlotDataItem):
            self.curve_plot_item = curve_plot_item
            self.curve_plot_item.setZValue(1)
        else:
            self.curve_plot_item = None

        if state is None:
            try:
                # Set the Tags list from the project configuration
                roi_defs = get_proj_config().options('ROI_DEFS')
                self._tags = dict.fromkeys(roi_defs)
            except NoProjectOpen as e:
                self._tags = {}
                # If a project isn't open
                warn('BaseROI subclass instance cannot load ROI_DEFS, no Project open.')
        else:
            # Restore states
            self._tags = state['tags']
            self.curve_data = state['curve_data']

        self.view_box = view_box
        self.roi_graphics_object = None
        self._color = None
        self._original_color = None

        self.metadata = metadata

    # @property
    # def curve_data(self) -> tuple:
    #     return self.curve_plot_item.getData()
    #
    # @curve_data.setter
    # def curve_data(self, data: tuple):
    #     """
    #
    #     :rtype: tuple
    #     """
    #     if self.curve_plot_item is None:
    #         return
    #     self.curve_plot_item.setData(x=data[0], y=data[1])

    @property
    def zValue(self) -> int:
        """ZValue of the curve plot graphics object and roi_graphics_object"""
        return self._zValue

    @zValue.setter
    def zValue(self, val: int):
        self._zValue = val

    def get_roi_graphics_object(self) -> QtWidgets.QGraphicsObject:
        pass

    def set_roi_graphics_object(self, *args, **kwargs):
        pass

    def reset_color(self):
        self.set_color(self._original_color)

    def set_original_color(self, color):
        self._original_color = color
        self.reset_color()

    def get_color(self):
        return self._color

    def set_color(self, color: Union[np.ndarray, str], *args, **kwargs):
        pen = pg.mkPen(color, *args, **kwargs)
        self.roi_graphics_object.setPen(pen)
        self.curve_plot_item.setPen(pen)
        if isinstance(self.roi_graphics_object, pg.ScatterPlotItem):
            self.roi_graphics_object.setBrush(pg.mkBrush(color, *args, **kwargs))
        self._color = color

    def set_text(self, text: str):
        text_item = pg.TextItem(text)
        # self.view_box.addItem()

    def set_tag(self, roi_def: str, tag: str):
        self._tags[roi_def] = tag

    def get_tag(self, roi_def) -> str:
        if self._tags[roi_def] is None:
            return ''
        return self._tags[roi_def]

    def get_all_tags(self) -> dict:
        d = deepcopy(self._tags)
        for key in d.keys():
            if d[key] is None:
                d[key] = ''
        return d

    def add_to_viewer(self):
        self.view_box.addItem(self.get_roi_graphics_object())

    def remove_from_viewer(self):
        roi = self.get_roi_graphics_object()
        self.view_box.removeItem(roi)
        if self.curve_plot_item is not None:
            self.curve_plot_item.clear()
        del self.curve_plot_item
        del roi

    def to_state(self):
        """Must be implemented in subclass"""
        raise NotImplementedError("Must be implemented in subclasses")

    @classmethod
    def from_state(cls, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, state: dict):
        raise NotImplementedError("Must be implemented in subclasses")


class ManualROI(BaseROI):
    """A class manually drawn ROIs"""
    def __init__(self, curve_plot_item: pg.PlotDataItem,
                 roi_graphics_object: pg.ROI,
                 view_box: pg.ViewBox,
                 state: Union[dict, None] = None,
                 spike_data: np.ndarray = None,
                 dfof_data: np.ndarray = None,):
        """
        :type state: dict
        """
        assert isinstance(roi_graphics_object, pg.ROI)

        super(ManualROI, self).__init__(curve_plot_item, view_box, state)

        self.set_roi_graphics_object(roi_graphics_object)

        if state is not None:
            self.add_to_viewer()
            self._set_roi_graphics_object_state(state['roi_graphics_object_state'])

    @property
    def curve_data(self) -> tuple:
        """
        :return: tuple of (xs, ys)
        """
        return self.curve_plot_item.getData()

    @curve_data.setter
    def curve_data(self, data: tuple):
        """
        Curve data as (xs, ys)
        """
        if self.curve_plot_item is None:
            return
        if data is None:
            self.curve_plot_item.clear()

        self.curve_plot_item.setData(x=data[0], y=data[1])

    def get_roi_graphics_object(self) -> pg.ROI:
        return self.roi_graphics_object

    def set_roi_graphics_object(self, graphics_object: pg.ROI):
        self.roi_graphics_object = graphics_object

    def _set_roi_graphics_object_state(self, state):
        self.roi_graphics_object.setState(state)

    def to_state(self):
        if isinstance(self.roi_graphics_object, pg.PolyLineROI):
            shape = 'PolyLineROI'
        elif isinstance(self.roi_graphics_object, pg.EllipseROI):
            shape = 'EllipseROI'

        state = {'curve_data': self.curve_data,
                 'shape': shape,
                 'roi_graphics_object_state': self.roi_graphics_object.saveState(),
                 'tags': self.get_all_tags(),
                 'roi_type': 'ManualROI'
                 }
        return state

    @staticmethod
    def get_generic_roi_graphics_object(shape: str, dims: tuple) -> pg.ROI:
        """Get a generic pg.ROI instance to add to the viewer. Its state can later be set from a saved state or it can be modified by the user"""
        x = dims[0]
        y = dims[1]

        if shape == 'PolyLineROI':
            roi_graphics_object = pg.PolyLineROI([[0, 0],
                                                  [int(0.1 * x), 0],
                                                  [int(0.1 * x), int(0.1 * y)],
                                                  [0, int(0.1 * y)]],
                                                 closed=True, pos=[0, 0], removable=True)
            return roi_graphics_object

        elif shape == 'EllipseROI':
            roi_graphics_object = pg.EllipseROI(pos=[0, 0], size=[int(x * 0.1), int(y * 0.1)], removable=True)
            return roi_graphics_object

    @classmethod
    def from_state(cls, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, state: dict):
        roi_graphics_object = ManualROI.get_generic_roi_graphics_object(state['shape'], (10, 10))
        return cls(curve_plot_item=curve_plot_item, roi_graphics_object=roi_graphics_object, view_box=view_box, state=state)

    @classmethod
    def from_positions(cls, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, positions: list):
        """
        From a list of positions. Used if importing ImageJ ROIs, can be used for other purposes as well.

        :param positions: list of (x, y) tuples; [(x1, y1), (x2, y2), ... (xn, yn)]
        """
        roi_graphics_object = pg.PolyLineROI(positions=positions, closed=True, removable=True)
        return cls(curve_plot_item=curve_plot_item, roi_graphics_object=roi_graphics_object, view_box=view_box)

    @classmethod
    def from_ellipse(cls, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, positions: list):
        """
        Create an ellipse from a list of positions when importing ImageJ ROIs.

        :param positions:  [origin_bottom_x, origin_bottom_y, width, height]
        """
        
        roi_graphics_object = pg.EllipseROI(pos=[positions[0], positions[1]], size=[int(positions[2]), int(positions[3])], removable=True)
        return cls(curve_plot_item=curve_plot_item, roi_graphics_object=roi_graphics_object, view_box=view_box)

class ScatterROI(BaseROI):
    """A class for unmoveable ROIs drawn using scatter points"""
    def __init__(self, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, state: Union[dict, None] = None,
                 curve_data: np.ndarray = None, xs: np.ndarray = None, ys: np.ndarray = None, metadata: dict = None,
                 spike_data: np.ndarray = None, dfof_data: np.ndarray = None,
                 **kwargs):
        """

        :param curve_plot_item:
        :param view_box:
        :param state:
        :param curve_data: 1D numpy array of y values
        :param kwargs:
        """
        self.spot_size = 1
        self.curve_data: List[np.ndarray, np.ndarray] = None  #: list of [xs, ys], xs and ys are 1D numpy arrays
        self._curve_data = None
        self.spike_data = None
        self.dfof_data = None

        super(ScatterROI, self).__init__(curve_plot_item, view_box, state, metadata=metadata)

        if (xs is not None) and (ys is not None):
            self.set_roi_graphics_object(xs, ys)

        if state is None:
            self.set_curve_data(curve_data)

            # set spikes
            if spike_data is not None:
                xs = np.arange(len(spike_data))
                self.spike_data = [xs, spike_data]

            # set dfof
            if dfof_data is not None:
                xs = np.arange(len(dfof_data))
                self.dfof_data = [xs, dfof_data]

        else:
            self.restore_state(state)

        self.curve_data_type = 'curve'

    # # override to get diff behavior from ManualROI
    # @property
    # def curve_data(self) -> tuple:
    #     return self._curve_data
    #
    # @curve_data.setter
    # def curve_data(self, y_vals: np.ndarray):
    #     xs = np.arange(len(y_vals))
    #     self._curve_data = [xs, y_vals]

    def set_curve_data(self, y_vals: np.ndarray, set_plot: bool = True):
        """Set the curve data"""
        xs = np.arange(len(y_vals))
        self.curve_data = [xs, y_vals]
        if set_plot:
            self.set_viewer_curveplot('curve')

    # sets the plot in the viewer
    def set_viewer_curveplot(self, data_type: str = 'curve'):
        """
        :param data_type: one of "curve", "spike", or "dfof"
        """
        if self.curve_plot_item is None:
            return

        data = getattr(self, f'{data_type}_data')
        if data is not None:
            self.curve_plot_item.setData(x=data[0], y=data[1])
        else:
            self.curve_plot_item.clear()
            warn(f'ROI does not contain the following '
                 f'data attribute: {data_type}_data')

    def restore_state(self, state: dict):
        if 'spike_data' in state.keys():
            self.spike_data = state['spike_data']
        if 'dfof_data' in state.keys():
            self.dfof_data = state['dfof_data']


        self.roi_xs = state['roi_xs']
        self.roi_ys = state['roi_ys']

        self._draw()

    def to_state(self) -> dict:
        state = {'roi_xs':      self.roi_xs,
                 'roi_ys':      self.roi_ys,
                 'curve_data':  self.curve_data,
                 'spike_data':  self.spike_data,
                 'dfof_data':   self.dfof_data,
                 'tags':        self.get_all_tags(),
                 'roi_type':    self.__class__.__name__,}

        return state

    def set_roi_graphics_object(self, xs: np.ndarray, ys: np.ndarray):
        self.roi_xs = xs.astype(int)
        self.roi_ys = ys.astype(int)

        self._draw()

    def get_roi_graphics_object(self) -> pg.ScatterPlotItem:
        if self.roi_graphics_object is None:
            raise AttributeError('Must call set_roi_graphics_object() first')

        return self.roi_graphics_object

    def _draw(self):
        """Create the scatter plot that is used for visualization of the spatial localization"""
        self.roi_graphics_object = pg.ScatterPlotItem(
            self.roi_xs, self.roi_ys, symbol='s', size=self.spot_size, pxMode=True
        )

    @classmethod
    def from_state(cls, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, state: dict, **kwargs):
        return cls(curve_plot_item=curve_plot_item, view_box=view_box, state=state, **kwargs)


class VolCNMF(ScatterROI):
    """3D ROI for CNMF data"""
    def __init__(self, curve_plot_item: pg.PlotDataItem	, view_box: pg.ViewBox, cnmf_idx: int = None,
                 curve_data: np.ndarray = None, contour: dict = None, state: Union[dict, None] = None,
                 spike_data: np.ndarray = None, dfof_data: np.ndarray = None, metadata: dict = None,
                 zlevel: int = 0):

        self.zlevel: int = zlevel  #: z-level of the ROI that is currently visible, different from zValue!!
        self.zcenter: int = None   #: z-level where this ROI has its center

        super(VolCNMF, self).__init__(curve_plot_item, view_box, state, curve_data,
                                      spike_data=spike_data, dfof_data=dfof_data,
                                      metadata=metadata)

        if state is None:
            # get the outline from the cnmf output
            self.coors = contour['coordinates']
            self.zcenter = int(contour['CoM'][0])

            self.set_roi_graphics_object()

            self.set_curve_data(curve_data)
            self.cnmf_idx = cnmf_idx  #: original index of the ROI from cnmf idx_components
        else:
            self.restore_state(state)

        self.visible = False

    def restore_state(self, state: dict):
        self.coors = state['coors']
        self.cnmf_idx = state['cnmf_idx']
        self.zcenter = state['zcenter']

        self.set_roi_graphics_object()

    def to_state(self) -> dict:
        state = {'curve_data':  self.curve_data,
                 'spike_data':  self.spike_data,
                 'dfof_data':   self.dfof_data,
                 'tags':        self.get_all_tags(),
                 'roi_type':    self.__class__.__name__,
                 'coors':       self.coors,
                 'cnmf_idx':    self.cnmf_idx,
                 'zcenter':     self.zcenter}

        return state

    def set_roi_graphics_object(self):
        coors = self.coors[self.zlevel]
        self.current_coors = coors[~np.isnan(coors).any(axis=1)]
        self.check_visible()

        self.roi_xs = self.current_coors[:, 1].flatten().astype(int)
        self.roi_ys = self.current_coors[:, 0].flatten().astype(int)

        self._draw()

    def check_visible(self):
        if self.current_coors.size > 0:
            self.visible = True
        else:
            self.visible = False

    def _draw(self):
        """Create the scatter plot that is used for visualization of the spatial localization"""
        self.roi_graphics_object = pg.ScatterPlotItem(
            self.roi_xs,
            self.roi_ys,
            symbol='s',
            size=self.spot_size
        )

    def set_zlevel(self, z: int):
        """
        Set the z-level of the ROI to correspond with the
        z-level of the image.

        Different from `setZValue`!!
        """

        self.zlevel = z

        coors = self.coors[self.zlevel]
        self.current_coors = coors[~np.isnan(coors).any(axis=1)]

        self.check_visible()

        self.roi_xs = self.current_coors[:, 1].flatten().astype(int)
        self.roi_ys = self.current_coors[:, 0].flatten().astype(int)

        self.set_viewer_curveplot(self.curve_data_type)

        self.roi_graphics_object.setData(self.roi_xs, self.roi_ys)#, symbol='s', size=self.spot_size)


class VolMultiCNMFROI(ScatterROI):
    """3D ROI for CNMF data created by runing CNMF independently on each plane"""
    def __init__(self, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, cnmf_idx: int = None,
                 curve_data: np.ndarray = None, contour: dict = None, state: Union[dict, None] = None,
                 spike_data: np.ndarray = None, dfof_data: np.ndarray = None, metadata: dict = None,
                 zlevel: int = 0, zcenter: int = None, parent_manager=None, roi_ix: int = None,
                 scatter_plot: pg.ScatterPlotItem = None, **kwargs):
        self.zlevel: int = zlevel  #: z-level of the ROI that is currently visible, different from zValue!!
        self.zcenter: int = zcenter  #: z-level where this ROI has its center, must be specified

        self.parent_manager = parent_manager

        self.roi_ix = roi_ix
        self.scatter_plot = scatter_plot

        if self.parent_manager is not None:
            ix = (self.parent_manager.roi_ixs[self.zcenter] == self.roi_ix).flatten()
            roi_xys = self.parent_manager.roi_xys[self.zcenter][ix]
            self.roi_xs = roi_xys[:, 0]
            self.roi_ys = roi_xys[:, 1]

        super(VolMultiCNMFROI, self).__init__(curve_plot_item, view_box, state, curve_data,
                                              spike_data=spike_data, dfof_data=dfof_data,
                                              metadata=metadata)

        if self.zcenter is None:
            raise ValueError(
                "zcenter must not be None"
            )

        if 'raw_min_max' in kwargs.keys():
            self.raw_min_max = kwargs.pop('raw_min_max')
        else:
            self.raw_min_max = None

        if state is None:
            self.set_curve_data(curve_data, set_plot=False)
            self.cnmf_idx = cnmf_idx  #: original index of the ROI from cnmf idx_components
        else:
            self.restore_state(state)

        self.check_visible()

    def get_roi_graphics_object(self) -> QtWidgets.QGraphicsObject:
        return self.roi_graphics_object

    def set_roi_graphics_object(self, *args, **kwargs):
        pass

    def get_color(self):
        pass

    def set_color(self, color: Union[np.ndarray, str], *args, **kwargs):
        # just assume white color for now
        colors = np.copy(self.parent_manager.roi_crs[self.zcenter])
        color = np.array([255., 255., 255., 255.])
        colors[(self.parent_manager.roi_ixs[self.zcenter] == self.roi_ix).flatten()] = color

        brushes = list(map(pg.mkBrush, colors))
        pens = list(map(pg.mkPen, colors))

        self.scatter_plot.setBrush(brushes)
        self.scatter_plot.setPen(pens)

        # self.curve_plot_item.setSize(pg.mkPen('w', width=2))

    def add_to_viewer(self):
        if self.parent_manager is None:
            super(VolMultiCNMFROI, self).add_to_viewer()

    def remove_from_viewer(self):
        if self.parent_manager is None:
            super(VolMultiCNMFROI, self).remove_from_viewer()
            return
        
        new_ixs = (self.parent_manager.roi_ixs[self.zcenter] != self.roi_ix).flatten()
        self.parent_manager.roi_ixs[self.zcenter] = self.parent_manager.roi_ixs[self.zcenter][new_ixs]
        self.parent_manager.roi_xys[self.zcenter] = self.parent_manager.roi_xys[self.zcenter][new_ixs]
        self.parent_manager.roi_crs[self.zcenter] = self.parent_manager.roi_crs[self.zcenter][new_ixs]

        xys = self.parent_manager.roi_xys[self.zcenter]
        crs = self.parent_manager.roi_crs[self.zcenter]

        self.scatter_plot.setData(
            xys[:, 0],
            xys[:, 1],
            symbol='s',
            size=1,
            pxMode=True,
            brush=list(map(pg.mkBrush, crs)),
            pen=list(map(pg.mkPen, crs))
        )

        if self.curve_plot_item is not None:
            self.curve_plot_item.clear()
        del self.curve_plot_item

    def restore_state(self, state):
        super(VolMultiCNMFROI, self).restore_state(state)
        # self.curve_data = state['curve_data']
        self.cnmf_idx = state['cnmf_idx']

        if 'raw_min' in state.keys() and 'raw_max' in state.keys():
            self.raw_min = state.pop('raw_min')
            self.raw_max = state.pop('raw_max')
        else:
            self.raw_min = None
            self.raw_max = None

        self.zcenter = state['zcenter']

        if self.parent_manager is None:
            self._draw()

    def to_state(self) -> dict:
        state = {'cnmf_idx': self.cnmf_idx,
                 'zcenter': self.zcenter,
                 'roi_ix': self.roi_ix,
                 'roi_xs': self.roi_xs,
                 'roi_ys': self.roi_ys,
                 'raw_min_max': self.raw_min_max,
                 'curve_data': self.curve_data,
                 'spike_data': self.spike_data,
                 'dfof_data': self.dfof_data,
                 'tags': self.get_all_tags(),
                 'roi_type': self.__class__.__name__,
                 }

        return state

    def check_visible(self):
        if self.zlevel == self.zcenter:
            self.visible = True
            self.set_viewer_curveplot(self.curve_data_type)
        else:
            self.visible = False
            self.curve_plot_item.clear()

    def set_zlevel(self, z: int):
        """
        Set the z-level of the ROI to correspond with the
        z-level of the image.

        Different from `setZValue`!!
        """

        self.zlevel = z
        self.check_visible()


class CNMFROI(ScatterROI):
    """A class for ROIs imported from CNMF(E) output data"""
    def __init__(self, curve_plot_item: pg.PlotDataItem, view_box: pg.ViewBox, cnmf_idx: int = None,
                 curve_data: np.ndarray = None, contour: dict = None, state: Union[dict, None] = None,
                 spike_data: np.ndarray = None, dfof_data: np.ndarray = None, metadata: dict = None,
                 **kwargs):
        """
        Instantiate attributes.

        :type: curve_data:  np.ndarray
        :param curve_data:  1D numpy array of y values
        :type  contour:     np.ndarray
        :type  state:       dict
        :param cnmf_idx:    original index of the ROI from cnmf idx_components
        """
        super(CNMFROI, self).__init__(curve_plot_item, view_box, state, curve_data,
                                      spike_data=spike_data, dfof_data=dfof_data,
                                      metadata=metadata)

        self.roi_xs = np.empty(0)  #: numpy array of the x values of the ROI's spatial coordinates
        self.roi_ys = np.empty(0)  #: numpy array of the y values of the ROI's spatial coordinates

        if 'raw_min_max' in kwargs.keys():
            self.raw_min_max = kwargs.pop('raw_min_max')
        else:
            self.raw_min_max = None

        if state is None:
            # get the outline from the cnmf output
            cors = contour['coordinates']
            cors = cors[~np.isnan(cors).any(axis=1)]

            xs = cors[:, 0].flatten()
            ys = cors[:, 1].flatten()

            self.set_roi_graphics_object(xs, ys)

            self.set_curve_data(curve_data)
            self.cnmf_idx = cnmf_idx  #: original index of the ROI from cnmf idx_components
        else:
            self.restore_state(state)

        # self.spot_size = 1

    def restore_state(self, state):
        super(CNMFROI, self).restore_state(state)
        # self.curve_data = state['curve_data']
        self.cnmf_idx = state['cnmf_idx']

        if 'raw_min' in state.keys() and 'raw_max' in state.keys():
            self.raw_min = state.pop('raw_min')
            self.raw_max = state.pop('raw_max')
        else:
            self.raw_min = None
            self.raw_max = None

    def to_state(self) -> dict:
        state = super(CNMFROI, self).to_state()

        state = {**state,
                 'cnmf_idx':    self.cnmf_idx,
                 'raw_min_max': self.raw_min_max
                 }

        return state



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 26 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from .link_viewers_dialog_pytemplate import *
from ...common import configuration
from ...pyqtgraphCore.imageview import ImageView


class ViewerLinkages(QtWidgets.QWidget):
    def __init__(self):
        QtWidgets.QWidget.__init__(self)

        self.ui = Ui_Form()
        self.ui.setupUi(self)

        self.linkages = {'source': None, 'receivers': list()}

        self.ui.btnSet.clicked.connect(self.slot_set)
        self.ui.btnDisconnect.clicked.connect(self.disconnect_all)

        self.ui.listWidgetReceivers.addItems(['0', '1', '2', '3', '4'])

    def populate_list_widgets(self):
        self.ui.listWidgetReceivers.clear()
        self.ui.listWidgetSource.clear()

        self.ui.listWidgetReceivers.addItems(list(range(len(configuration.window_manager.viewers))))
        self.ui.listWidgetSource.addItems(list(range(len(configuration.window_manager.viewers))))

    def disconnect_all(self):
        source = self.linkages['source']
        if source is None:
            return
        assert isinstance(source, ImageView)
        for receiver in self.linkages['receivers']:
            assert isinstance(receiver, ImageView)
            source.sigTimeChanged.disconnect(receiver.setCurrentIndex)

        self.linkages = {'source': None, 'receivers': list()}

    def slot_set(self):
        self.disconnect_all()
        src_ix = self.ui.listWidgetSource.currentRow()
        self.ui.listWidgetSource.currentItem().setBackground(QtGui.QBrush(QtGui.QColor('yellow')))
        source = configuration.window_manager.viewers[src_ix]

        self.linkages['source'] = source

        for item in self.ui.listWidgetReceivers.items():
            item.setBackground(QtGui.QBrush(QtGui.QColor('yellow')))

        qmodel_ixs = self.ui.listWidgetReceivers.selectedIndexes()
        receiver_ixs = [ix.row() for ix in qmodel_ixs]

        self.linkages['receivers'] = [configuration.window_manager.viewers[ix] for ix in receiver_ixs]

        self.connect_receivers(self.linkages['source'], self.linkages['receivers'])

    def connect_receivers(self, source: ImageView, receivers: list):
        for receiver in receivers:
            assert isinstance(receiver, ImageView)
            source.sigTimeChanged.connect(receiver.setCurrentIndex)


if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    w = ViewerLinkages()
    w.show()
    app.exec_()


# -*- coding: utf-8 -*-
from PyQt5 import QtWidgets
from sklearn import manifold
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn import decomposition
from .common import *
from ....analysis import organize_dataframe_columns
import pandas as pd
from ....plotting import NeuralDecomposePlot


class Manifold(CtrlNode):
    """Manifold learning"""
    nodeName = 'Manifold'
    uiTemplate = [('data_column', 'combo', {}),

                  ('method', 'combo', {'items': ['Isomap', 'Locally Linear Embedding',
                                                 'Spectral Embedding', 'MDS']}),

                  ('n_components', 'intSpin', {'max': 3, 'min': 2, 'value': 2, 'step': 1}),

                  ('kwargs', 'kwargTextEdit', {'placeHolder': 'Any additional kwargs to pass to'
                                                              'the sklearn.manifold function'}),

                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()

        method = self.ctrls['method'].currentText()
        n_components = self.ctrls['n_components'].value()
        kwargs = self.ctrls['kwargs'].get_kwargs()

        output_column = '_MANIFOLD'

        data = np.vstack(self.t.df[self.data_column].values)

        T = self._get_transform(method, data, n_components, kwargs)
        self.t.df[output_column] = T.tolist()

        params = {'data_column': self.data_column,
                  'method': method,
                  'n_components': n_components,
                  'kwargs': kwargs
                  }

        self.t.history_trace.add_operation(data_block_id='all', operation='manifold', parameters=params)
        self.t.last_output = output_column

        return self.t

    def _get_transform(self, method: str, data: np.ndarray, n_components: int, kwargs: dict):
        if method == 'Isomap':
            return manifold.Isomap(n_components=n_components, **kwargs).fit_transform(data)

        elif method == 'Locally Linear Embedding':
            return manifold.LocallyLinearEmbedding(n_components=n_components, **kwargs).fit_transform(data)

        elif method == 'Spectral Embedding':
            return manifold.SpectralEmbedding(n_components=n_components, **kwargs).fit_transform(data)

        elif method == 'MDS':
            return manifold.MDS(n_components=n_components, **kwargs).fit_transform(data)


class LDA(CtrlNode):
    """Linear Discriminant Analysis, uses sklearn"""
    nodeName = "LDA"
    uiTemplate = [('train_data', 'list_widget', {'selection_mode': QtWidgets.QAbstractItemView.ExtendedSelection,
                                                 'toolTip': 'Column containing the training data'}),
                  ('train_labels', 'combo', {'toolTip': 'Column containing training labels'}),
                  ('solver', 'combo', {'items': ['svd', 'lsqr', 'eigen']}),
                  ('shrinkage', 'combo', {'items': ['None', 'auto', 'value']}),
                  ('shrinkage_val', 'doubleSpin', {'min': 0.0, 'max': 1.0, 'step': 0.1, 'value': 0.5}),
                  ('n_components', 'intSpin', {'min': 2, 'max': 1000, 'step': 1, 'value': 2}),
                  ('tol', 'intSpin', {'min': -50, 'max': 0, 'step': 1, 'value': -4}),
                  ('score', 'lineEdit', {}),
                  ('predict_on', 'list_widget', {'selection_mode': QtWidgets.QAbstractItemView.ExtendedSelection,
                                                 'toolTip': 'Data column of the input "predict" Transmission\n'
                                                            'that is used for predicting from the model'}),
                  ('Apply', 'check', {'applyBox': True, 'checked': False})
                  ]

    def __init__(self, name, **kwargs):
        CtrlNode.__init__(self, name, terminals={'train': {'io': 'in'},
                                                 'predict': {'io': 'in'},

                                                 'T': {'io': 'out'},
                                                 'coef': {'io': 'out'},
                                                 'means': {'io': 'out'},
                                                 'predicted': {'io': 'out'}
                                                 },
                          **kwargs)
        self.ctrls['score'].setReadOnly(True)

    def process(self, **kwargs):
        return self.processData(**kwargs)

    def processData(self, train: Transmission, predict: Transmission):
        self.t = train.copy()  #: Transmisison instance containing the training data with the labels
        if predict is not None:
            self.to_predict = predict.copy()  #: Transmission instance containing the data to predict after fitting on the the training data

        dcols, ccols, ucols = organize_dataframe_columns(self.t.df.columns)

        self.ctrls['train_data'].setItems(dcols)
        self.ctrls['train_labels'].setItems(ccols)

        if predict is not None:
            pdcols, ccols, ucols = organize_dataframe_columns(self.to_predict.df.columns)
            self.ctrls['predict_on'].setItems(pdcols)

        if not self.apply_checked():
            return

        train_columns = self.ctrls['train_data'].getSelectedItems()
        labels = self.ctrls['train_labels'].currentText()

        solver = self.ctrls['solver'].currentText()

        shrinkage = self.ctrls['shrinkage'].currentText()
        if shrinkage == 'value':
            shrinkage = self.ctrls['shrinkage_val'].value()
        elif shrinkage == 'None':
            shrinkage = None

        n_components = self.ctrls['n_components'].value()
        tol = 10 ** self.ctrls['tol'].value()

        store_covariance = True if solver == 'svd' else False

        params = {'train_data': train_columns,
                  'train_labels': labels,
                  'solver': solver,
                  'shrinkage': shrinkage,
                  'n_components': n_components,
                  'tol': tol,
                  'store_covariance': store_covariance
                  }

        kwargs = params.copy()
        kwargs.pop('train_data')
        kwargs.pop('train_labels')
        self.lda = LinearDiscriminantAnalysis(**kwargs)

        # Make an array of all the data from the selected columns
        self.X = np.hstack([np.vstack(self.t.df[train_column]) for train_column in train_columns])
        self.y = self.t.df[labels]

        self.X_ = self.lda.fit_transform(self.X, self.y)

        self.t.df['_LDA_TRANSFORM'] = self.X_.tolist()
        self.t.df['_LDA_TRANSFORM'] = self.t.df['_LDA_TRANSFORM'].apply(np.array)

        params.update({'score': self.lda.score(self.X, self.y),
                       'classes': self.lda.classes_.tolist()
                       })

        self.ctrls['score'].setText(f"{params['score']:.4f}")

        self.t.history_trace.add_operation('all', 'lda', params)

        self.t.df['_LDA_DFUNC'] = self.lda.decision_function(self.X).tolist()

        coef_df = pd.DataFrame({'classes': self.lda.classes_, '_COEF': self.lda.coef_.tolist()})
        t_coef = Transmission(df=coef_df, history_trace=self.t.history_trace)

        means_df = pd.DataFrame({'classes': self.lda.classes_, '_MEANS': self.lda.means_.tolist()})
        t_means = Transmission(df=means_df, history_trace=self.t.history_trace)

        out = {'T': self.t, 'coef': t_coef, 'means': t_means, 'predicted': None}

        # Predict using the trained model
        predict_columns = self.ctrls['predict_on'].getSelectedItems()

        if not predict_columns:
            return out

        if predict_columns != train_columns:
            QtWidgets.QMessageBox.warning('Predict and Train columns do not match',
                                          'The selected train and predict columns are different')

        predict_data = np.hstack([np.vstack(self.to_predict.df[predict_column]) for predict_column in predict_columns])
        self.to_predict.df['LDA_PREDICTED_LABELS'] = self.lda.predict(predict_data)
        self.to_predict.df['_LDA_TRANSFORM'] = self.lda.transform(predict_data).tolist()
        self.to_predict.df['_LDA_TRANSFORM'] = self.to_predict.df['_LDA_TRANSFORM'].apply(np.array)

        params_predict = params.copy()
        params_predict.update({'predict_columns': predict_columns})

        self.to_predict.history_trace.add_operation('all', 'lda-predict', params_predict)

        out.update({'predicted': self.to_predict})

        return out


class PCA(CtrlNode):
    """Principal component analysis. Uses sklearn.decomposition.PCA"""
    nodeName = 'PCA'
    uiTemplate = [('data_column', 'combo', {}),

                  ('n_components', 'intSpin', {'min': 2, 'max': 999, 'value': 2, 'step': 1}),

                  ('whiten', 'check', {'chcked': False}),

                  ('svd_solver', 'combo', {'items': ['auto', 'full', 'arpack', 'randomized']}),

                  ('tol', 'doubleSpin', {'min': 0.0, 'max': 999.0, 'value': 0, 'step': 0.05}),

                  ('iterated_power', 'intSpin', {'min': -1, 'max': 999, 'value': -1, 'step': 1,
                                                 'toolTip': 'If value is -1 then iterated power is set to "auto"'}),
                  ('exp_var', 'lineEdit', {'readOnly': True, 'toolTip': 'Variance explained by each component'}),

                  ('exp_var_p', 'lineEdit', {'readOnly': True,
                                             'toolTip': 'Percentage of variance explained by each component'}),

                  ('sing_vals', 'lineEdit', {'readOnly': True, 'toolTip': 'Singular values for each component'}),

                  ('Apply', 'check', {'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()

        output_column = '_PCA_TRANSFORM'

        n_components = self.ctrls['n_components'].value()
        whiten = self.ctrls['whiten'].isChecked()
        svd_solver = self.ctrls['svd_solver'].currentText()
        tol = self.ctrls['tol'].value()
        iterated_power = self.ctrls['iterated_power'].value()
        if iterated_power == -1:
            iterated_power = 'auto'

        params = {'n_components': n_components,
                  'whiten': whiten,
                  'svd_solver': svd_solver,
                  'tol': tol,
                  iterated_power: iterated_power
                  }

        pca = decomposition.PCA(n_components=n_components, whiten=whiten, svd_solver=svd_solver, tol=tol,
                                iterated_power=iterated_power)

        self.X = np.vstack(self.t.df[self.data_column].values)

        self.X_ = pca.fit_transform(self.X)

        exp_var = pca.explained_variance_
        exp_var_p = pca.explained_variance_ratio_

        params.update({**params,
                       'exp_var': exp_var,
                       'exp_var_p': exp_var_p
                       })

        self.t.df[output_column] = self.X_.tolist()
        self.t.df[output_column] = self.t.df['_PCA_TRANSFORM'].apply(np.array)

        self.t.history_trace.add_operation('all', 'pca', params)

        self.ctrls['exp_var'].setText(str(exp_var))
        self.ctrls['exp_var_p'].setText(str(exp_var_p))
        self.ctrls['sing_vals'].setText(str(pca.singular_values_))

        return self.t


class NeuralDynamics(CtrlNode):
    """Dimensionality reduction of neuronal dynamics"""
    nodeName = 'NeuralDynamics'
    uiTemplate = [('ShowGUI', 'button', {'text': 'Show GUI'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.widget = NeuralDecomposePlot()
        self.ctrls['ShowGUI'].clicked.connect(self.widget.show)

    def process(self, output_transmission=False, **kwargs):
        self.processData(**kwargs)

    def processData(self, **kwargs):
        self.transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(self.transmissions)
        self.t = Transmission.merge(self.transmissions_list)
        self.widget.set_input(self.t)
        return None



# -*- coding: utf-8 -*-
from ... import PlotDataItem, ComboBox
from ....analysis import simple_plot_window
from .common import *
import numpy as np
import pandas as pd
from ....plotting.widgets import HeatmapTracerWidget, ScatterPlotWidget, \
    BeeswarmPlotWindow, ProportionsWidget, CrossCorrelationWidget, SpaceMapWidget
from ....plotting.variants.timeseries import TimeseriesPlot
from graphviz import Digraph
from ....common.utils import make_workdir
import os
import subprocess
from random import randrange


class Plot(CtrlNode):
    """Plot curves and/or scatter points"""
    nodeName = 'Plot'
    uiTemplate = [('Show', 'check', {'checked': True}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True}),
                  ('data_column', 'combo', {})
                  ]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})#, 'Out': {'io': 'out'}})
        self.trans_ids = []
        self.plot_widget = simple_plot_window.PlotWindow()
        self.ctrls['Apply'].clicked.connect(self.update)
        self.ctrls['Show'].clicked.connect(self.plot_widget.setVisible)

    def process(self, **kwargs):
        transmissions = kwargs['In']

        if not len(transmissions) > 0:
            raise Exception('No incoming transmissions')

        transmissions = merge_transmissions(transmissions)

        columns = pd.concat([t.df for t in transmissions]).columns
        dcols = organize_dataframe_columns(columns)[0]
        self.ctrls['data_column'].setItems(dcols)

        if self.ctrls['Apply'].isChecked() is False:
            return
        data_column = self.ctrls['data_column'].currentText()
        self.plot_widget.graphicsView.clear()

        srcs = []
        plots = []
        colors = ['m', 'r', 'y', 'g', 'c', 'b']
        ci = 0
        for t in transmissions:
            if t is None:
                srcs.append(str(type(None)))
                continue

            if id(t) in self.trans_ids:
                continue

            for ix, r in t.df.iterrows():

                if data_column in r.index:
                    data = r[data_column]
                # elif 'peak_curve' in r.index:
                #     data = r['peak_curve']
                else:
                    srcs.append('No data found in chosen column!')
                    continue
                if data is None:
                    srcs.append('No data found in chosen column!')
                    continue

                plot = PlotDataItem()
                try:
                    plot.setData(data)
                    # plot.setData(data / np.maximum(np.min(data), 0.0000000001))
                    plot.setPen(colors[ci])
                except Exception as e:
                    srcs.append('Plotting error: ' + str(e))
                    continue

                self.plot_widget.graphicsView.addItem(plot)

            ci += 1

        self.plot_widget.set_history_widget(srcs)


class FrequencyDomainMagnitude(CtrlNode):
    """Plot Frequency vs. Frequency Domain Magnitude"""
    nodeName = 'FDM Plot'
    uiTemplate = [('Show', 'check', {'checked': True}),
                  ('Apply', 'check', {'checked': True, 'applyBox': True})
                  ]


class Timeseries(CtrlNode):
    """Seaborn Timeseries plot"""
    nodeName = 'Timeseries'
    uiTemplate = [('data_column', 'combo', {}),
                  ('err_style', 'combo', {'items': ['band', 'bars']}),
                  ('Show', 'button', {'text': 'Show'}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True}),
                  ]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.ctrls['Apply'].clicked.connect(self.update)
        self.plot_widget = TimeseriesPlot()
        self.ctrls['Show'].clicked.connect(self.plot_widget.show)

    def process(self, **kwargs):
        transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(transmissions)

        self.t = Transmission.merge(self.transmissions_list)

        columns = self.t.df.columns.to_list()

        self.ctrls['data_column'].setItems(columns)

        if not self.ctrls['Apply'].isChecked():
            return

        data_column = self.ctrls['data_column'].currentText()
        data = np.vstack(self.t.df[data_column].values)

        es = self.ctrls['err_style'].currentText()

        self.plot_widget.set(data=data, err_style=es)


class Heatmap(CtrlNode):
    """Stack 1-D arrays and plot visually like a heatmap"""
    nodeName = "Heatmap"
    uiTemplate = [('Show', 'button', {'text': 'Show GUI'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})#, 'Out': {'io': 'out'}})
        self.plot_widget = HeatmapTracerWidget()
        self.ctrls['Show'].clicked.connect(self.plot_widget.show)

    def process(self, **kwargs):
        self.transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(self.transmissions)
        if len(self.transmissions_list) == 1:
            self.t = self.transmissions_list[0]
        else:
            self.t = Transmission.merge(self.transmissions_list)

        self.plot_widget.set_input(self.t)


class ScatterPlot(CtrlNode):
    """Scatter Plot, useful for visualizing transformed data and clusters"""
    nodeName = "ScatterPlot"
    uiTemplate = [('Show', 'button', {'text': 'Show'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.plot_widget = ScatterPlotWidget()
        self.ctrls['Show'].clicked.connect(self.plot_widget.show)

    def process(self, **kwargs):
        transmissions = kwargs['In']

        tran_list = merge_transmissions(transmissions)
        merged = Transmission.merge(tran_list)

        self.plot_widget.set_input(merged)


class BeeswarmPlots(CtrlNode):
    """Beeswarm and Violin plots"""
    nodeName = 'BeeswarmPlots'
    uiTemplate = [('Apply', 'check', {'checked': False, 'applyBox': True}),
                  ('ShowGUI', 'button', {'text': 'OpenGUI'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.plot_widget = None
        self.ctrls['ShowGUI'].clicked.connect(self._open_plot_gui)

    def process(self, **kwargs):
        if (self.ctrls['Apply'].isChecked() is False) or self.plot_widget is None:
            return

        transmissions = kwargs['In']
        transmissions_list = merge_transmissions(transmissions)

        self.plot_widget.update_input_transmissions(transmissions_list)

    def _open_plot_gui(self):
        if self.plot_widget is None:
            self.plot_widget = BeeswarmPlotWindow(parent=None)
        self.plot_widget.show()


class Proportions(CtrlNode):
    """Plot proportions of one categorical column vs another"""
    nodeName = 'Proportions'
    uiTemplate = [('show gui', 'button', {'text': 'Show Gui'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.plot_widget = ProportionsWidget()
        self.ctrls['show gui'].clicked.connect(self.plot_widget.show)

    def process(self, **kwargs):
        self.transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(self.transmissions)

        self.t = Transmission.merge(self.transmissions_list)

        self.plot_widget.set_input(self.t)


class CrossCorr(CtrlNode):
    """Cross Correlation"""
    nodeName = 'CrossCorr'
    uiTemplate = [('ShowGUI', 'button', {'text': 'Show Gui'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.plot_widget = CrossCorrelationWidget()
        self.ctrls['ShowGUI'].clicked.connect(self.plot_widget.show)

    def process(self, **kwargs):
        self.transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(self.transmissions)

        self.t = Transmission.merge(self.transmissions_list)

        self.plot_widget.set_input(self.t)


class SpaceMap(CtrlNode):
    """Visualize spatial maps of a categorical variable"""
    nodeName = 'SpaceMap'
    uiTemplate = [('ShowGUI', 'button', {'text': 'Show GUI'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.plot_widget =SpaceMapWidget()
        self.ctrls['ShowGUI'].clicked.connect(self.plot_widget.show)

    def process(self, **kwargs):
        self.transmissions = kwargs['In']
        self.transmissions = merge_transmissions(self.transmissions)

        self.t = Transmission.merge(self.transmissions)

        self.plot_widget.set_input(self.t)


class AnalysisGraph(CtrlNode):
    """Graph of the analysis log"""
    nodeName = 'AnalysisGraph'
    uiTemplate = [('data_blocks', 'list_widget', {'toolTip': 'Double click a data block to open the pdf'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.ctrls['data_blocks'].itemDoubleClicked.connect(self.open_file)

    def process(self, **kwargs):
        self.transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(self.transmissions)

        self.t = Transmission.merge(self.transmissions_list)

        workdir = make_workdir(f'analysis_graphs{randrange(1000, 9999)}')

        self.fnames = []

        for db in self.t.history_trace.data_blocks:
            fname = os.path.join(workdir, f"{db}")
            g = Digraph('G', filename=fname, format='pdf', node_attr={'shape': 'record'})

            db_history = self.t.history_trace.get_data_block_history(db)

            o_encounters = []
            o_n_l = []

            for op in db_history:
                o = list(op.keys())[0]
                params = op[o]
                params = cleanup_params(o, params)
                s = []
                for k in params.keys():
                    p = str(params[k]).replace("{", "\\n   --\> ").replace("}", "").replace("|", "\|")  # .replace("[", " ").replace("(", " ").replace(")", " ")
                    s.append(f"{k}: {p}")

                o_n = f"{o}.{o_encounters.count(o)}"
                o_n_l.append(o_n)
                s = '{ ' + o_n + ' | ' + '\\n'.join(s) + ' }'

                g.node(o_n, s)
                o_encounters.append(o)

            es = []
            for i in range(1, len(o_n_l)):
                es.append((o_n_l[i - 1], o_n_l[i]))

            g.edges(es)
            g.render(view=False)

            self.fnames.append(fname + '.pdf')

        self.ctrls['data_blocks'].setItems(list(map(str, range(len(self.fnames)))))

    def open_file(self, item):
        ix = self.ctrls['data_blocks'].currentRow()
        subprocess.call(["xdg-open", self.fnames[ix]])


def cleanup_params(operation, params):
    log = params.copy()
    if operation == 'rfft':
        log.pop('frequencies')
    elif operation =='fcluster':
        log.pop('linkage_matrix')

    return log




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2019 Kushal Kolar
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from fcsugar import *
import numpy as np
from scipy import signal as scipy_signal
from ..containers import TransmissionContainer


@node
def sort_peak_widths(container: TransmissionContainer, data_column: str):
    peaks = container.df[data_column]
    widths = [scipy_signal.peak_widths(p, [np.argmax(p)])[0][0] for p in peaks]

    width_sorter = np.argsort(widths)

    container.df = container.df.reindex(index=width_sorter).reset_index(drop=True)

    return container



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './roi_manager_pytemplate.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_DockWidget(object):
    def setupUi(self, DockWidget):
        DockWidget.setObjectName("DockWidget")
        DockWidget.resize(384, 566)
        DockWidget.setMinimumSize(QtCore.QSize(384, 566))
        DockWidget.setFloating(True)
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tabWidget = QtWidgets.QTabWidget(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy)
        self.tabWidget.setObjectName("tabWidget")
        self.tabStandard = QtWidgets.QWidget()
        self.tabStandard.setObjectName("tabStandard")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tabStandard)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_3 = QtWidgets.QLabel(self.tabStandard)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(self.tabStandard)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout.addWidget(self.label_4)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.horizontalSliderSpotSize = QtWidgets.QSlider(self.tabStandard)
        self.horizontalSliderSpotSize.setMinimum(1)
        self.horizontalSliderSpotSize.setMaximum(20)
        self.horizontalSliderSpotSize.setPageStep(1)
        self.horizontalSliderSpotSize.setProperty("value", 1)
        self.horizontalSliderSpotSize.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalSliderSpotSize.setObjectName("horizontalSliderSpotSize")
        self.verticalLayout_3.addWidget(self.horizontalSliderSpotSize)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.btnAddROI = QtWidgets.QPushButton(self.tabStandard)
        self.btnAddROI.setObjectName("btnAddROI")
        self.horizontalLayout_3.addWidget(self.btnAddROI)
        self.checkBoxShowAll = QtWidgets.QCheckBox(self.tabStandard)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.checkBoxShowAll.sizePolicy().hasHeightForWidth())
        self.checkBoxShowAll.setSizePolicy(sizePolicy)
        self.checkBoxShowAll.setObjectName("checkBoxShowAll")
        self.horizontalLayout_3.addWidget(self.checkBoxShowAll)
        self.checkBoxLivePlot = QtWidgets.QCheckBox(self.tabStandard)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.checkBoxLivePlot.sizePolicy().hasHeightForWidth())
        self.checkBoxLivePlot.setSizePolicy(sizePolicy)
        self.checkBoxLivePlot.setObjectName("checkBoxLivePlot")
        self.horizontalLayout_3.addWidget(self.checkBoxLivePlot)
        self.btnPlot = QtWidgets.QPushButton(self.tabStandard)
        self.btnPlot.setObjectName("btnPlot")
        self.horizontalLayout_3.addWidget(self.btnPlot)
        self.verticalLayout_3.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_5 = QtWidgets.QLabel(self.tabStandard)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_2.addWidget(self.label_5)
        self.radioButton_curve_data = QtWidgets.QRadioButton(self.tabStandard)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.radioButton_curve_data.sizePolicy().hasHeightForWidth())
        self.radioButton_curve_data.setSizePolicy(sizePolicy)
        self.radioButton_curve_data.setObjectName("radioButton_curve_data")
        self.horizontalLayout_2.addWidget(self.radioButton_curve_data)
        self.radioButton_dfof = QtWidgets.QRadioButton(self.tabStandard)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.radioButton_dfof.sizePolicy().hasHeightForWidth())
        self.radioButton_dfof.setSizePolicy(sizePolicy)
        self.radioButton_dfof.setObjectName("radioButton_dfof")
        self.horizontalLayout_2.addWidget(self.radioButton_dfof)
        self.radioButton_spikes = QtWidgets.QRadioButton(self.tabStandard)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.radioButton_spikes.sizePolicy().hasHeightForWidth())
        self.radioButton_spikes.setSizePolicy(sizePolicy)
        self.radioButton_spikes.setObjectName("radioButton_spikes")
        self.horizontalLayout_2.addWidget(self.radioButton_spikes)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.verticalLayout_3.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.pushButtonImportFromImageJ = QtWidgets.QPushButton(self.tabStandard)
        self.pushButtonImportFromImageJ.setObjectName("pushButtonImportFromImageJ")
        self.horizontalLayout_4.addWidget(self.pushButtonImportFromImageJ)
        self.btnSwitchToManualMode = QtWidgets.QPushButton(self.tabStandard)
        self.btnSwitchToManualMode.setEnabled(False)
        self.btnSwitchToManualMode.setObjectName("btnSwitchToManualMode")
        self.horizontalLayout_4.addWidget(self.btnSwitchToManualMode)
        self.verticalLayout_3.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.tabStandard)
        self.label.setMaximumSize(QtCore.QSize(16777215, 75))
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.listWidgetROIs = QtWidgets.QListWidget(self.tabStandard)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.listWidgetROIs.sizePolicy().hasHeightForWidth())
        self.listWidgetROIs.setSizePolicy(sizePolicy)
        self.listWidgetROIs.setMinimumSize(QtCore.QSize(50, 0))
        self.listWidgetROIs.setMaximumSize(QtCore.QSize(80, 16777215))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.listWidgetROIs.setFont(font)
        self.listWidgetROIs.setObjectName("listWidgetROIs")
        self.verticalLayout.addWidget(self.listWidgetROIs)
        self.horizontalLayout_5.addLayout(self.verticalLayout)
        self.verticalLayoutTags = QtWidgets.QVBoxLayout()
        self.verticalLayoutTags.setObjectName("verticalLayoutTags")
        self.label_2 = QtWidgets.QLabel(self.tabStandard)
        self.label_2.setAlignment(QtCore.Qt.AlignCenter)
        self.label_2.setObjectName("label_2")
        self.verticalLayoutTags.addWidget(self.label_2)
        self.listWidgetROITags = QtWidgets.QListWidget(self.tabStandard)
        self.listWidgetROITags.setObjectName("listWidgetROITags")
        self.verticalLayoutTags.addWidget(self.listWidgetROITags)
        self.horizontalLayout_5.addLayout(self.verticalLayoutTags)
        self.verticalLayout_3.addLayout(self.horizontalLayout_5)
        self.lineEditROITag = ROITagLineEdit(self.tabStandard)
        self.lineEditROITag.setObjectName("lineEditROITag")
        self.verticalLayout_3.addWidget(self.lineEditROITag)
        self.btnSetROITag = QtWidgets.QPushButton(self.tabStandard)
        self.btnSetROITag.setObjectName("btnSetROITag")
        self.verticalLayout_3.addWidget(self.btnSetROITag)
        self.tabWidget.addTab(self.tabStandard, "")
        self.tabMetaROI = QtWidgets.QWidget()
        self.tabMetaROI.setObjectName("tabMetaROI")
        self.tabWidget.addTab(self.tabMetaROI, "")
        self.verticalLayout_2.addWidget(self.tabWidget)
        DockWidget.setWidget(self.dockWidgetContents)

        self.retranslateUi(DockWidget)
        self.tabWidget.setCurrentIndex(0)
        self.lineEditROITag.returnPressed.connect(self.btnSetROITag.click)
        self.horizontalSliderSpotSize.valueChanged['int'].connect(self.label_4.setNum)
        QtCore.QMetaObject.connectSlotsByName(DockWidget)
        DockWidget.setTabOrder(self.tabWidget, self.horizontalSliderSpotSize)
        DockWidget.setTabOrder(self.horizontalSliderSpotSize, self.btnAddROI)
        DockWidget.setTabOrder(self.btnAddROI, self.checkBoxShowAll)
        DockWidget.setTabOrder(self.checkBoxShowAll, self.checkBoxLivePlot)
        DockWidget.setTabOrder(self.checkBoxLivePlot, self.btnPlot)
        DockWidget.setTabOrder(self.btnPlot, self.pushButtonImportFromImageJ)
        DockWidget.setTabOrder(self.pushButtonImportFromImageJ, self.btnSwitchToManualMode)
        DockWidget.setTabOrder(self.btnSwitchToManualMode, self.listWidgetROIs)
        DockWidget.setTabOrder(self.listWidgetROIs, self.listWidgetROITags)
        DockWidget.setTabOrder(self.listWidgetROITags, self.lineEditROITag)
        DockWidget.setTabOrder(self.lineEditROITag, self.btnSetROITag)

    def retranslateUi(self, DockWidget):
        _translate = QtCore.QCoreApplication.translate
        DockWidget.setWindowTitle(_translate("DockWidget", "&ROI Manager"))
        self.label_3.setText(_translate("DockWidget", "Spot size:"))
        self.label_4.setText(_translate("DockWidget", "1"))
        self.btnAddROI.setText(_translate("DockWidget", "Add ROI"))
        self.checkBoxShowAll.setText(_translate("DockWidget", "Show all"))
        self.checkBoxLivePlot.setText(_translate("DockWidget", "Live plot"))
        self.btnPlot.setText(_translate("DockWidget", "Plot"))
        self.label_5.setText(_translate("DockWidget", "View:"))
        self.radioButton_curve_data.setToolTip(_translate("DockWidget", "Plot the curve data"))
        self.radioButton_curve_data.setText(_translate("DockWidget", "cur&ve data"))
        self.radioButton_dfof.setToolTip(_translate("DockWidget", "Plot ΔF/F (if available)"))
        self.radioButton_dfof.setText(_translate("DockWidget", "ΔF/F"))
        self.radioButton_spikes.setToolTip(_translate("DockWidget", "Plot spikes (if available)"))
        self.radioButton_spikes.setText(_translate("DockWidget", "spi&kes"))
        self.pushButtonImportFromImageJ.setText(_translate("DockWidget", "Import from ImageJ"))
        self.btnSwitchToManualMode.setText(_translate("DockWidget", "Switch to manual mode"))
        self.label.setText(_translate("DockWidget", "ROIs"))
        self.label_2.setText(_translate("DockWidget", "Tags"))
        self.lineEditROITag.setPlaceholderText(_translate("DockWidget", "Add Tag to ROI Definition"))
        self.btnSetROITag.setText(_translate("DockWidget", "Set ROI tag"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabStandard), _translate("DockWidget", "Standard ROIs"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabMetaROI), _translate("DockWidget", "Meta ROIs"))

from ..roi_manager_modules.roi_tag_line_edit import ROITagLineEdit



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 17 17:59:04 2017

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

A simple class that does the back-end handilng to export image sequence data.

"""

import numpy as np
import cv2
import tifffile
from functools import partial
from .exporter_pytemplate import *


# vmin & vmax only has to be supplied for video outputs since videos cannot contain
# the full range of 16astype(np.int32 bit image depth (OpenCV & 16 bit grayscale is a nightmare to deal with)
class Exporter:
    def __init__(self, img_data, out_file, **kwargs):
        if out_file[-4:] != 'tiff' and 'levels' not in kwargs.keys():
            raise ValueError('levels must be specified for file formats other than tiff or npz')

        if 'fps' in kwargs.keys():
            self.fps = kwargs['fps']
        else:
            self.fps = img_data.meta['fps']

        self.img_data = img_data
        self.out_file = out_file
        
        if 'levels' in kwargs.keys():
            vmin = kwargs['levels'][0]
            vmax = kwargs['levels'][1]
            self.vmin = vmin
            self.vmax = vmax
            
        outFormat = self.out_file[-4:]

        formats = {'.tif': self.tiff,
                   'tiff': self.tiff,
                   '.npz': self.npz,
                   '.avi': partial(self.vidWrite, 'MJPG'),
                   '.mp4': partial(self.vidWrite, 'X264'),
                   '.gif': self.gif
                   }

        formats[outFormat]()

    def gif(self):
        scaled = self.scale_levels()
        with imageio.get_writer(self.out_file, mode='I', fps=self.fps) as writer:
            for frame in range(0, scaled.shape[2]):
                writer.append_data(scaled[:, :, frame].T)

    def tiff(self):
        tifffile.imsave(self.out_file, self.img_data.T)

    def npz(self):
        print('yay')

    def scale_levels(self):

        scaled = ((self.img_data - self.vmin) * (255/self.vmax)).clip(min=0, max=255).astype(np.uint8)
#        scaled = (((self.img_data.seq - self.vmin) * (self.vmax - self.vmin)) * (255.0/self.vmax)).clip(min=0, max=255).astype(np.uint8)#.clip(min=0, max=255).astype(np.uint8)
#        self.min_substract = (self.img_data - self.vmin).astype(np.uint8)
        return scaled

    def vidWrite(self, codec):
        fourcc = cv2.VideoWriter_fourcc(*codec)
        print('YAY CODEC IS ' + codec)
        out = cv2.VideoWriter(self.out_file, fourcc, int(self.fps),
                        (self.img_data.shape[0],
                         self.img_data.shape[1]))
        scaled = self.scale_levels()
        # if (self.vmin and self.vmax) == None:
        #     self.vmin = self.img_data.vmin
        #     self.vmax = self.img_data.vmax
        # self.min_substract = self.img_data.seq.astype(np.int32) - self.vmin
        # self.scaled = ((self.img_data.seq.astype(np.int32) - self.vmin)
        #                 * (255/self.vmax)).clip(min=0, max=255).astype(np.uint8)
#        self.no_neg = self.min_substract.clip(min=0)
#        self.scaled = np.array((self.no_neg * (255/self.vmax)), dtype=np.uint8)
        for frame in range(0, scaled.shape[2]):
            #print('entered write loop, frame# ' + str(frame))
            out.write(cv2.cvtColor(scaled[:, :, frame].T, cv2.COLOR_GRAY2BGR))
        out.release()


class ExporterGUI(QtWidgets.QWidget, Ui_exporter_template):
    def __init__(self):
        super().__init__()
        self.setupUi(self)
        self.show()
        formats = ['tiff', 'MJPG', 'X264', 'gif']
        self.comboBoxFormat.addItems(formats)
        self.labelSlider.setText(str(self.sliderFPS_Scaling.value()/10))
        self.sliderFPS_Scaling.valueChanged.connect(lambda v: self.labelSlider.setText(str(v/10)))
        self.comboBoxFormat.currentTextChanged.connect(self._enable_non_tiff)

    def _enable_non_tiff(self, f):
        if f == 'tiff':
            self._enable_non_tiff_ui(False)
        else:
            self._enable_non_tiff_ui(True)

    def _enable_non_tiff_ui(self, b=True):
        self.radioAuto.setEnabled(b)
        self.radioFromViewer.setEnabled(b)
        self.sliderFPS_Scaling.setEnabled(True)
        # self.checkBoxPseudocolor.setChecked(b)
        self.checkBoxPseudocolor.setEnabled(b)

    def file_path_dialog(self):
        path = QtWidgets.QFileDialog.getSaveFileName(self, 'Export filename, do not type an extension', '', '(*.*)')
        if path == '':
            return

        self.lineEdPath.setText(path[0])


from .common import *
import numpy as np
from ....analysis.math import drfft_dtw


class DRFFT_DTW(CtrlNode):
    """Compute distances using Dyanamic Time Warping between: 1)Raw curve and 2) The inverse discrete fourier transform
    derived from feeding incremental steps of frequency domain from discrete fourier transform of the raw curve."""
    nodeName = 'DRFFT_DTW'
    uiTemplate= [('data_column', 'combo',
                  {'toolTip': 'Data column to use as the "raw curve", such as spliced arrays of dF/Fo'}),

                 ('step_size', 'intSpin',
                  {'max': 5000, 'min': 10, 'value': 100, 'step': 50,
                   'toolTip': 'Step size for incrementing rfft inputs for computing irfft. Reducing step size will '
                              'reduce the computation time.'}),

                 ('step_start', 'intSpin',
                  {'max': 5000, 'min': 10, 'value': 100, 'step': 50,
                   'toolTip': 'Frequency domain to start incrementing steps from'}),

                 ('interpolate', 'check', {'toolTip': 'Interpoate the inverse fourier transforms. Increases computation'
                                                      'time SIGNIFICANTLY but may produce more accurate results.'}),

                 ('start', 'button', {'text': 'Compute'})
                 # ('Apply', 'check', {'checked': False, 'applyBox': True})
                 ]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})#, 'Out': {'io': 'out', 'bypass': 'In'}})
        self.ctrls['start'].clicked.connect(self._compute)
        self.results = None
        self.widget = drfft_dtw.Widget()

    def process(self, **kwargs):
        transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(transmissions)

        self.t = Transmission.merge(self.transmissions_list).copy()
        columns = self.t.df.columns.to_list()
        self.ctrls['data_column'].setItems(columns)

        self.param_step = self.ctrls['step_size'].value()
        self.param_start = self.ctrls['step_start'].value()
        self.param_interpolate = self.ctrls['interpolate'].isChecked()

        self.data = np.vstack(self.t.df[self.data_column])

    def _compute(self):
        # self.ctrls['start'].setDisabled(True)
        self.widget.show()
        start = self.param_start
        step = self.param_step
        interp = self.param_interpolate
        self.widget.start(start, step, interp, self.data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on April 23 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ..core.common import ViewerUtils
from .pytemplates.cnmfe_pytemplate import *
import json
from ...common import get_window_manager
from ...common.qdialogs import *
from uuid import UUID
from copy import deepcopy


class ModuleGUI(QtWidgets.QDockWidget):
    def __init__(self, parent, viewer_reference):
        self.vi = ViewerUtils(viewer_reference)

        QtWidgets.QDockWidget.__init__(self, parent)

        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)

        self.ui.btnAddToBatchCorrPNR.clicked.connect(lambda: self.add_to_batch_corr_pnr())
        self.ui.btnAddToBatchCNMFE.clicked.connect(lambda: self.add_to_batch_cnmfe())

    @present_exceptions()
    def get_params(self, item_type: str, group_params: bool = False) -> dict:
        """
        Get a dict of the set parameters.
        If the work environment was loaded from a motion correction batch item it put the bord_px in the dict.
        Doesn't use any arguments

        :param item_type: one of `corr_pnr` or `cnmfe`
        """
        if self.vi.viewer.workEnv.imgdata.meta['fps'] == 0:
            raise KeyError('No framerate for current image sequence!',
                           'You must set a framerate for the current image sequence. '
                           'You can set it manually in the console like this:\nget_meta()["fps"] = <framerate>')

        # Get bord_px param if motion correction was performed
        history_trace = self.vi.viewer.workEnv.history_trace
        try:
            bord_px = next(d for ix, d in enumerate(history_trace) if 'caiman_motion_correction' in d)['caiman_motion_correction']['bord_px']
        except StopIteration:
            bord_px = 0

        deconv = self.ui.comboBoxDeconv.currentText()
        if deconv == 'SKIP':
            method_deconvolution = None
            deconv_flag = False
        else:
            deconv_flag = True
            method_deconvolution = deconv

        gSig = self.ui.spinBoxGSig.value()

        # Correlation-PNR param
        corr_pnr_kwargs = \
            {
                'gSig': gSig
            }

        kval = self.ui.spinBoxK.value()
        if kval == 0:
            kval = None

        # CNMFE kwargs
        cnmfe_kwargs = \
            {
                'gSig': (gSig, gSig),
                'gSiz': (4 * gSig + 1, 4 * gSig + 1),
                'border_pix': bord_px,
                'rf': self.ui.spinBoxRf.value(),
                'stride': self.ui.spinBoxOverlap.value(),
                'gnb': self.ui.spinBoxGnb.value(),
                'nb_patch': self.ui.spinBoxNb_patch.value(),
                'K': kval,
                'min_corr': self.ui.doubleSpinBoxMinCorr.value(),
                'min_pnr': self.ui.spinBoxMinPNR.value(),
                'deconv_flag': deconv_flag,
                'method_deconvolution': method_deconvolution,
                'Ain': self.ui.lineEditAin.text().replace(' ', '').strip('\n'),
                'p': self.ui.spinBox_p.value(),
                'merge_thresh': self.ui.doubleSpinBoxMergeThresh.value(),
                'ssub': self.ui.spinBox_ssub.value(),
                'tsub': self.ui.spinBox_tsub.value(),
                'low_rank_background': self.ui.checkBox_low_rank_background.isChecked(),
                'ring_size_factor': self.ui.doubleSpinBox_ring_size_factor.value(),
                'update_background_components': True,
                'del_duplicates': True,
                'fr': self.vi.viewer.workEnv.imgdata.meta['fps']
            }

        # Any additional CNMFE kwargs set in the text entry
        if self.ui.groupBox_cnmf_kwargs.isChecked():
            try:
                _kwargs = self.ui.plainTextEdit_cnmf_kwargs.toPlainText()
                cnmfe_kwargs_add = eval(f"dict({_kwargs})")
                cnmfe_kwargs.update(cnmfe_kwargs_add)
            except:
                raise ValueError("CNMFE kwargs not formatted properly.")

        # kwargs for component evaluation
        eval_kwargs = \
            {
                'fr': self.vi.viewer.workEnv.imgdata.meta['fps'],
                'min_SNR': self.ui.spinBoxMinSNR.value(),
                'rval_thr': self.ui.doubleSpinBoxRValuesMin.value(),
                'decay_time': self.ui.doubleSpinBoxDecayTime.value(),
            }

        # Any additional eval kwargs set in the text entry
        if self.ui.groupBox_eval_kwargs.isChecked():
            try:
                _kwargs = self.ui.plainTextEdit_eval_kwargs.toPlainText()
                eval_kwargs_add = eval(f"dict({_kwargs})")
                eval_kwargs.update(eval_kwargs_add)
            except:
                raise ValueError("Evaluation kwargs not formatted properly.")

        # Output either the dict for corr-pnr or CNMFE
        if item_type == 'corr_pnr':
            item_name = self.ui.lineEdCorrPNRName.text()
            d = \
                {
                    'item_name':        item_name,
                    'do_cnmfe':         False,
                    'do_corr_pnr':      True,
                    'border_pix':       bord_px
                }

            if group_params:
                d.update({'corr_pnr_kwargs': corr_pnr_kwargs})
            else:
                d.update({**corr_pnr_kwargs})

        elif item_type == 'cnmfe':
            item_name = self.ui.lineEdName.text()
            d = \
                {
                    'item_name':        item_name,
                    'do_cnmfe':         True,
                    'do_corr_pnr':      False,
                    'border_pix':       bord_px
                }
            if group_params:
                d.update(
                    {
                        'cnmfe_kwargs': cnmfe_kwargs,
                        'eval_kwargs': eval_kwargs,
                    }
                )
            else:
                d.update(
                    {
                        **cnmfe_kwargs,
                        **eval_kwargs
                    }
                )
        else:
            raise ValueError("Must specify argument `item_type` as either 'corr_pnr' or 'cnmfe'")

        d['keep_memmap'] = self.ui.checkBoxKeepMemmap.isChecked()

        return d

    def add_to_batch_corr_pnr(self, params: dict = None) -> UUID:
        """
        Add a Corr PNR batch item with the currently set parameters and the current work environment.

        """
        if self.vi.viewer.workEnv.isEmpty:
            QtWidgets.QMessageBox.warning(self, 'Empty work environment', 'The work environment is empty, '
                                                                          'nothing to add to batch')
            return
        input_workEnv = self.vi.viewer.workEnv

        if params is None:
            d = self.get_params('corr_pnr', group_params=True)
        else:
            # Check that user passed dict is formatted correctly
            required_keys = ['item_name', 'do_cnmfe', 'do_corr_pnr', 'border_pix', 'corr_pnr_kwargs']
            if any([k not in params.keys() for k in required_keys]):
                raise ValueError(f'Must pass a params dict with the following keys:\n'
                                 f'{required_keys}\n'
                                 f'Please see the docs for more information.')
            d = deepcopy(params)

        batch_manager = get_window_manager().get_batch_manager()
        name = d['item_name']

        self.vi.viewer.status_bar_label.showMessage('Please wait, adding Corr PNR: ' + name + ' to batch...')

        u = batch_manager.add_item(module='CNMFE',
                                   name=name,
                                   input_workEnv=input_workEnv,
                                   input_params=d,
                                   info=d
                                   )

        self.vi.viewer.status_bar_label.showMessage('Done adding Corr PNR: ' + name + ' to batch!')
        self.clear_line_edits()

        return u

    def add_to_batch_cnmfe(self, params: dict = None) -> UUID:
        """
        Add a CNMFE batch item with the currently set parameters and the current work environment.
        """
        if self.vi.viewer.workEnv.isEmpty:
            QtWidgets.QMessageBox.warning(self, 'Empty work environment', 'The work environment is empty, '
                                                                          'nothing to add to batch')
            return
        input_workEnv = self.vi.viewer.workEnv

        if params is None:
            d = self.get_params('cnmfe', group_params=True)
        else:
            # Check that user passed dict is formatted correctly
            required_keys = ['item_name', 'do_cnmfe', 'do_corr_pnr', 'border_pix', 'cnmfe_kwargs', 'eval_kwargs']
            if any([k not in params.keys() for k in required_keys]):
                raise ValueError(f'Must pass a params dict with the following keys:\n'
                                 f'{required_keys}\n'
                                 f'Please see the docs for more information.')
            d = deepcopy(params)

        name = d['item_name']
        self.vi.viewer.status_bar_label.showMessage('Please wait, adding CNMFE: ' + name + ' to batch...')

        batch_manager = get_window_manager().get_batch_manager()
        u = batch_manager.add_item(module='CNMFE',
                                   name=name,
                                   input_workEnv=input_workEnv,
                                   input_params=d,
                                   info=d
                                   )

        if u is None:
            self.vi.viewer.status_bar_label.clearMessage()
            return

        self.vi.viewer.status_bar_label.showMessage('Done adding CNMFE: ' + name + ' to batch!')
        self.clear_line_edits()

        return u

    def clear_line_edits(self):
        self.ui.lineEdCorrPNRName.clear()
        self.ui.lineEdName.clear()
        self.ui.lineEditAin.clear()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './control_widget.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_CrossCorrelationControls(object):
    def setupUi(self, CrossCorrelationControls):
        CrossCorrelationControls.setObjectName("CrossCorrelationControls")
        CrossCorrelationControls.resize(904, 706)
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(CrossCorrelationControls)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.splitter = QtWidgets.QSplitter(CrossCorrelationControls)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayoutLeftMain = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayoutLeftMain.setContentsMargins(0, 0, 0, 0)
        self.verticalLayoutLeftMain.setObjectName("verticalLayoutLeftMain")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout()
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.label_5 = QtWidgets.QLabel(self.layoutWidget)
        font = QtGui.QFont()
        font.setPointSize(14)
        font.setBold(True)
        font.setWeight(75)
        self.label_5.setFont(font)
        self.label_5.setAlignment(QtCore.Qt.AlignCenter)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_8.addWidget(self.label_5)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_10 = QtWidgets.QLabel(self.layoutWidget)
        self.label_10.setObjectName("label_10")
        self.verticalLayout_6.addWidget(self.label_10)
        self.comboBoxDataColumn = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBoxDataColumn.setMinimumSize(QtCore.QSize(160, 0))
        self.comboBoxDataColumn.setObjectName("comboBoxDataColumn")
        self.verticalLayout_6.addWidget(self.comboBoxDataColumn)
        self.label_12 = QtWidgets.QLabel(self.layoutWidget)
        self.label_12.setObjectName("label_12")
        self.verticalLayout_6.addWidget(self.label_12)
        self.comboBoxLabelsColumn = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBoxLabelsColumn.setObjectName("comboBoxLabelsColumn")
        self.verticalLayout_6.addWidget(self.comboBoxLabelsColumn)
        self.horizontalLayout_2.addLayout(self.verticalLayout_6)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_2 = QtWidgets.QLabel(self.layoutWidget)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.radioButtonLag = QtWidgets.QRadioButton(self.layoutWidget)
        self.radioButtonLag.setChecked(True)
        self.radioButtonLag.setObjectName("radioButtonLag")
        self.verticalLayout_2.addWidget(self.radioButtonLag)
        self.radioButtonMaxima = QtWidgets.QRadioButton(self.layoutWidget)
        self.radioButtonMaxima.setObjectName("radioButtonMaxima")
        self.verticalLayout_2.addWidget(self.radioButtonMaxima)
        self.horizontalLayout_2.addLayout(self.verticalLayout_2)
        self.verticalLayout_8.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3.addLayout(self.verticalLayout_8)
        self.verticalLayout_7 = QtWidgets.QVBoxLayout()
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.label_11 = QtWidgets.QLabel(self.layoutWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_11.setFont(font)
        self.label_11.setAlignment(QtCore.Qt.AlignCenter)
        self.label_11.setObjectName("label_11")
        self.verticalLayout_7.addWidget(self.label_11)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(self.layoutWidget)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.checkBoxAbsoluteValue = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxAbsoluteValue.setChecked(True)
        self.checkBoxAbsoluteValue.setObjectName("checkBoxAbsoluteValue")
        self.verticalLayout_3.addWidget(self.checkBoxAbsoluteValue)
        self.doubleSpinBoxLagThreshold = QtWidgets.QDoubleSpinBox(self.layoutWidget)
        self.doubleSpinBoxLagThreshold.setMinimum(-999.0)
        self.doubleSpinBoxLagThreshold.setMaximum(999.0)
        self.doubleSpinBoxLagThreshold.setSingleStep(0.5)
        self.doubleSpinBoxLagThreshold.setProperty("value", 1.0)
        self.doubleSpinBoxLagThreshold.setObjectName("doubleSpinBoxLagThreshold")
        self.verticalLayout_3.addWidget(self.doubleSpinBoxLagThreshold)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_4 = QtWidgets.QLabel(self.layoutWidget)
        self.label_4.setAlignment(QtCore.Qt.AlignCenter)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_4.addWidget(self.label_4)
        self.doubleSpinBoxMaximaThreshold = QtWidgets.QDoubleSpinBox(self.layoutWidget)
        self.doubleSpinBoxMaximaThreshold.setMinimum(-1.0)
        self.doubleSpinBoxMaximaThreshold.setMaximum(1.0)
        self.doubleSpinBoxMaximaThreshold.setSingleStep(0.01)
        self.doubleSpinBoxMaximaThreshold.setProperty("value", 0.5)
        self.doubleSpinBoxMaximaThreshold.setObjectName("doubleSpinBoxMaximaThreshold")
        self.verticalLayout_4.addWidget(self.doubleSpinBoxMaximaThreshold)
        self.horizontalLayout.addLayout(self.verticalLayout_4)
        self.verticalLayout_7.addLayout(self.horizontalLayout)
        self.label_13 = QtWidgets.QLabel(self.layoutWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_13.setFont(font)
        self.label_13.setObjectName("label_13")
        self.verticalLayout_7.addWidget(self.label_13)
        self.comboBoxStimulusType = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBoxStimulusType.setObjectName("comboBoxStimulusType")
        self.verticalLayout_7.addWidget(self.comboBoxStimulusType)
        self.horizontalLayout_3.addLayout(self.verticalLayout_7)
        self.verticalLayoutLeftMain.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.pushButtonComputeAllData = QtWidgets.QPushButton(self.layoutWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.pushButtonComputeAllData.setFont(font)
        self.pushButtonComputeAllData.setObjectName("pushButtonComputeAllData")
        self.horizontalLayout_6.addWidget(self.pushButtonComputeAllData)
        self.pushButtonExportAllData = QtWidgets.QPushButton(self.layoutWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.pushButtonExportAllData.setFont(font)
        self.pushButtonExportAllData.setObjectName("pushButtonExportAllData")
        self.horizontalLayout_6.addWidget(self.pushButtonExportAllData)
        self.verticalLayoutLeftMain.addLayout(self.horizontalLayout_6)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_6 = QtWidgets.QLabel(self.layoutWidget)
        font = QtGui.QFont()
        font.setPointSize(14)
        font.setBold(True)
        font.setWeight(75)
        self.label_6.setFont(font)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout_5.addWidget(self.label_6)
        self.lineEditCurve1UUID = QtWidgets.QLineEdit(self.layoutWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lineEditCurve1UUID.setFont(font)
        self.lineEditCurve1UUID.setReadOnly(True)
        self.lineEditCurve1UUID.setObjectName("lineEditCurve1UUID")
        self.horizontalLayout_5.addWidget(self.lineEditCurve1UUID)
        self.verticalLayoutLeftMain.addLayout(self.horizontalLayout_5)
        self.graphicsViewCurve1 = PlotWidget(self.layoutWidget)
        self.graphicsViewCurve1.setObjectName("graphicsViewCurve1")
        self.verticalLayoutLeftMain.addWidget(self.graphicsViewCurve1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_7 = QtWidgets.QLabel(self.layoutWidget)
        font = QtGui.QFont()
        font.setPointSize(14)
        font.setBold(True)
        font.setWeight(75)
        self.label_7.setFont(font)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_4.addWidget(self.label_7)
        self.lineEditCurve2UUID = QtWidgets.QLineEdit(self.layoutWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lineEditCurve2UUID.setFont(font)
        self.lineEditCurve2UUID.setReadOnly(True)
        self.lineEditCurve2UUID.setObjectName("lineEditCurve2UUID")
        self.horizontalLayout_4.addWidget(self.lineEditCurve2UUID)
        self.verticalLayoutLeftMain.addLayout(self.horizontalLayout_4)
        self.graphicsViewCurve2 = PlotWidget(self.layoutWidget)
        self.graphicsViewCurve2.setObjectName("graphicsViewCurve2")
        self.verticalLayoutLeftMain.addWidget(self.graphicsViewCurve2)
        self.label_8 = QtWidgets.QLabel(self.layoutWidget)
        font = QtGui.QFont()
        font.setPointSize(14)
        font.setBold(True)
        font.setWeight(75)
        self.label_8.setFont(font)
        self.label_8.setObjectName("label_8")
        self.verticalLayoutLeftMain.addWidget(self.label_8)
        self.layoutWidget1 = QtWidgets.QWidget(self.splitter)
        self.layoutWidget1.setObjectName("layoutWidget1")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_9 = QtWidgets.QLabel(self.layoutWidget1)
        font = QtGui.QFont()
        font.setPointSize(14)
        font.setBold(True)
        font.setWeight(75)
        self.label_9.setFont(font)
        self.label_9.setObjectName("label_9")
        self.verticalLayout_5.addWidget(self.label_9)
        self.listWidgetSampleID = QtWidgets.QListWidget(self.layoutWidget1)
        self.listWidgetSampleID.setObjectName("listWidgetSampleID")
        self.verticalLayout_5.addWidget(self.listWidgetSampleID)
        self.verticalLayout_9.addWidget(self.splitter)

        self.retranslateUi(CrossCorrelationControls)
        QtCore.QMetaObject.connectSlotsByName(CrossCorrelationControls)
        CrossCorrelationControls.setTabOrder(self.comboBoxLabelsColumn, self.radioButtonLag)
        CrossCorrelationControls.setTabOrder(self.radioButtonLag, self.radioButtonMaxima)
        CrossCorrelationControls.setTabOrder(self.radioButtonMaxima, self.checkBoxAbsoluteValue)
        CrossCorrelationControls.setTabOrder(self.checkBoxAbsoluteValue, self.doubleSpinBoxLagThreshold)
        CrossCorrelationControls.setTabOrder(self.doubleSpinBoxLagThreshold, self.doubleSpinBoxMaximaThreshold)
        CrossCorrelationControls.setTabOrder(self.doubleSpinBoxMaximaThreshold, self.comboBoxStimulusType)
        CrossCorrelationControls.setTabOrder(self.comboBoxStimulusType, self.pushButtonComputeAllData)
        CrossCorrelationControls.setTabOrder(self.pushButtonComputeAllData, self.pushButtonExportAllData)
        CrossCorrelationControls.setTabOrder(self.pushButtonExportAllData, self.lineEditCurve1UUID)
        CrossCorrelationControls.setTabOrder(self.lineEditCurve1UUID, self.graphicsViewCurve1)
        CrossCorrelationControls.setTabOrder(self.graphicsViewCurve1, self.lineEditCurve2UUID)
        CrossCorrelationControls.setTabOrder(self.lineEditCurve2UUID, self.graphicsViewCurve2)
        CrossCorrelationControls.setTabOrder(self.graphicsViewCurve2, self.listWidgetSampleID)
        CrossCorrelationControls.setTabOrder(self.listWidgetSampleID, self.comboBoxDataColumn)

    def retranslateUi(self, CrossCorrelationControls):
        _translate = QtCore.QCoreApplication.translate
        CrossCorrelationControls.setWindowTitle(_translate("CrossCorrelationControls", "Form"))
        self.label_5.setText(_translate("CrossCorrelationControls", "Parameters"))
        self.label_10.setText(_translate("CrossCorrelationControls", "Data column:"))
        self.label_12.setText(_translate("CrossCorrelationControls", "Labels column:"))
        self.label_2.setText(_translate("CrossCorrelationControls", "Visualization:"))
        self.radioButtonLag.setText(_translate("CrossCorrelationControls", "&lag"))
        self.radioButtonMaxima.setText(_translate("CrossCorrelationControls", "&maxima"))
        self.label_11.setText(_translate("CrossCorrelationControls", "Thresholds"))
        self.label_3.setText(_translate("CrossCorrelationControls", "lag"))
        self.checkBoxAbsoluteValue.setText(_translate("CrossCorrelationControls", "absolute value"))
        self.label_4.setText(_translate("CrossCorrelationControls", "maxima"))
        self.label_13.setText(_translate("CrossCorrelationControls", "Stimulus Type:"))
        self.pushButtonComputeAllData.setText(_translate("CrossCorrelationControls", "Compute all data"))
        self.pushButtonExportAllData.setText(_translate("CrossCorrelationControls", "Export all data"))
        self.label_6.setText(_translate("CrossCorrelationControls", "Curve 1"))
        self.label_7.setText(_translate("CrossCorrelationControls", "Curve 2"))
        self.label_8.setText(_translate("CrossCorrelationControls", "Cross-correlation function"))
        self.label_9.setText(_translate("CrossCorrelationControls", "SampleID"))

from ....pyqtgraphCore import PlotWidget



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './exporter_pytemplate.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_exporter_template(object):
    def setupUi(self, exporter_template):
        exporter_template.setObjectName("exporter_template")
        exporter_template.resize(302, 304)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(exporter_template)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(exporter_template)
        self.label.setMaximumSize(QtCore.QSize(75, 16777215))
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.comboBoxFormat = QtWidgets.QComboBox(exporter_template)
        self.comboBoxFormat.setObjectName("comboBoxFormat")
        self.horizontalLayout.addWidget(self.comboBoxFormat)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem)
        self.labelSlider = QtWidgets.QLabel(exporter_template)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelSlider.sizePolicy().hasHeightForWidth())
        self.labelSlider.setSizePolicy(sizePolicy)
        self.labelSlider.setAlignment(QtCore.Qt.AlignCenter)
        self.labelSlider.setObjectName("labelSlider")
        self.horizontalLayout_5.addWidget(self.labelSlider)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem1)
        self.verticalLayout_2.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_4 = QtWidgets.QLabel(exporter_template)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_2.addWidget(self.label_4)
        self.sliderFPS_Scaling = QtWidgets.QSlider(exporter_template)
        self.sliderFPS_Scaling.setEnabled(False)
        self.sliderFPS_Scaling.setMinimum(1)
        self.sliderFPS_Scaling.setMaximum(100)
        self.sliderFPS_Scaling.setSingleStep(5)
        self.sliderFPS_Scaling.setProperty("value", 10)
        self.sliderFPS_Scaling.setOrientation(QtCore.Qt.Horizontal)
        self.sliderFPS_Scaling.setObjectName("sliderFPS_Scaling")
        self.horizontalLayout_2.addWidget(self.sliderFPS_Scaling)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_3 = QtWidgets.QLabel(exporter_template)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_3.addWidget(self.label_3)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.radioAuto = QtWidgets.QRadioButton(exporter_template)
        self.radioAuto.setEnabled(False)
        self.radioAuto.setChecked(True)
        self.radioAuto.setObjectName("radioAuto")
        self.verticalLayout.addWidget(self.radioAuto)
        self.radioFromViewer = QtWidgets.QRadioButton(exporter_template)
        self.radioFromViewer.setEnabled(False)
        self.radioFromViewer.setObjectName("radioFromViewer")
        self.verticalLayout.addWidget(self.radioFromViewer)
        self.horizontalLayout_3.addLayout(self.verticalLayout)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.checkBoxPseudocolor = QtWidgets.QCheckBox(exporter_template)
        self.checkBoxPseudocolor.setObjectName("checkBoxPseudocolor")
        self.verticalLayout_2.addWidget(self.checkBoxPseudocolor)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.lineEdPath = QtWidgets.QLineEdit(exporter_template)
        self.lineEdPath.setObjectName("lineEdPath")
        self.horizontalLayout_4.addWidget(self.lineEdPath)
        self.btnChoosePath = QtWidgets.QPushButton(exporter_template)
        self.btnChoosePath.setMaximumSize(QtCore.QSize(31, 16777215))
        self.btnChoosePath.setObjectName("btnChoosePath")
        self.horizontalLayout_4.addWidget(self.btnChoosePath)
        self.verticalLayout_2.addLayout(self.horizontalLayout_4)
        self.progressBar = QtWidgets.QProgressBar(exporter_template)
        self.progressBar.setProperty("value", 24)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_2.addWidget(self.progressBar)
        self.btnExport = QtWidgets.QPushButton(exporter_template)
        self.btnExport.setObjectName("btnExport")
        self.verticalLayout_2.addWidget(self.btnExport)

        self.retranslateUi(exporter_template)
        QtCore.QMetaObject.connectSlotsByName(exporter_template)
        exporter_template.setTabOrder(self.comboBoxFormat, self.sliderFPS_Scaling)
        exporter_template.setTabOrder(self.sliderFPS_Scaling, self.radioAuto)
        exporter_template.setTabOrder(self.radioAuto, self.radioFromViewer)
        exporter_template.setTabOrder(self.radioFromViewer, self.lineEdPath)
        exporter_template.setTabOrder(self.lineEdPath, self.btnChoosePath)

    def retranslateUi(self, exporter_template):
        _translate = QtCore.QCoreApplication.translate
        exporter_template.setWindowTitle(_translate("exporter_template", "Form"))
        self.label.setText(_translate("exporter_template", "Format"))
        self.labelSlider.setText(_translate("exporter_template", "TextLabel"))
        self.label_4.setText(_translate("exporter_template", "fps scaling"))
        self.sliderFPS_Scaling.setToolTip(_translate("exporter_template", "Speed up or slow down the framerate (only for video and gifs)"))
        self.label_3.setToolTip(_translate("exporter_template", "Lookup table min & max, only for non-tiff files."))
        self.label_3.setText(_translate("exporter_template", "LUT limits:"))
        self.radioAuto.setToolTip(_translate("exporter_template", "From meta-data (if any)"))
        self.radioAuto.setText(_translate("exporter_template", "Auto"))
        self.radioFromViewer.setToolTip(_translate("exporter_template", "From the current min & max as set in the viewer window"))
        self.radioFromViewer.setText(_translate("exporter_template", "From &viewer"))
        self.checkBoxPseudocolor.setText(_translate("exporter_template", "Pseudocolor from Viewer"))
        self.lineEdPath.setPlaceholderText(_translate("exporter_template", "path"))
        self.btnChoosePath.setText(_translate("exporter_template", "..."))
        self.btnExport.setText(_translate("exporter_template", "Export"))




from .heatmap import Heatmap
from .lineplot import Lineplot
from .violins import ViolinsPlot
from .timeseries import TimeseriesPlot
from .pgscatter import PgScatterPlot
from .pg_curve_plot import PgCurvePlot
from .beeswarms import BeeswarmPlot

__all__ = ['Heatmap', 'Lineplot', 'ViolinsPlot', 'TimeseriesPlot', 'PgScatterPlot', 'PgCurvePlot', 'BeeswarmPlot']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon March 5 2017

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt
from typing import Union


class HistoryTreeWidget(QtWidgets.QTreeWidget):
    _exclude_top_level_expansion = {'CurvePath', 'ImgInfoPath', 'ImgPath', 'ImgUUID',
                                    'ROI_State','meta', 'sitm_maps', 'uuid_curve'}
    
    def __init__(self, parent=None):
        super(HistoryTreeWidget, self).__init__(parent)
        self.top_level_keys = None

    def fill_widget(self, src: Union[dict, list]):
        """
        :param src: dict or list to fill in the tree widget.
        """
        self.clear()
        if type(src) is dict:
            self.top_level_keys = set(src.keys()) - self._exclude_top_level_expansion

        self._fill_item(self.invisibleRootItem(), src)

    def _fill_item(self, item, value):
        tn = 0
        # item.setExpanded(True)
        if type(value) is dict:
            for key, val in sorted(value.items()):
                child = QtWidgets.QTreeWidgetItem()
                child.setText(0, str(key))
                item.addChild(child)
                self._fill_item(child, val)

        elif type(value) is list:
            for val in value:
                child = QtWidgets.QTreeWidgetItem()
                item.addChild(child)
                if type(val) is dict:
                    child.setText(0, self._get_key(val))
                    for v in val.values():
                        self._fill_item(child, v)
                elif type(val) is list:
                    child.setText(0, 'Transmission ' + str(tn))
                    self._fill_item(child, val)
                    tn += 1
                else:
                    child.setText(0, str(val))
                # child.setExpanded(False)
        else:
            child = QtWidgets.QTreeWidgetItem()
            child.setText(0, str(value))
            item.addChild(child)

        if self.top_level_keys is not None:
            self._expand_top_level()

    def _expand_top_level(self):
        for k in self.top_level_keys:
            if str(k).startswith('_'):
                continue
            items = self.findItems(k, Qt.MatchExactly)
            for item in items:
                item.setExpanded(True)

    def _get_key(self, d):
        k = ''
        for key in d.keys():
            k = key + k
        return k
    
if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    win = QtWidgets.QMainWindow()
    history_tree = HistoryTreeWidget()
    win.setCentralWidget(history_tree)
    import pickle

    df = pickle.load(open('/share/data/longterm/4/kushal/kushal_on_ssd2/cell_types_project/fft_data_with_index.trn', 'rb'))['df']

    history_tree.fill_widget(df.iloc[0].to_dict())
    win.resize(300, 600)

    win.show()
    app.exec_()



from PyQt5.QtGui import QFont, QSyntaxHighlighter, QTextCharFormat, QColor
from PyQt5.QtWidgets import QApplication, QMainWindow, QTextEdit
from pygments import highlight
from pygments.lexers.python import Python3Lexer
from pygments.formatter import Formatter
from pygments.token import Name, Keyword


def hex2QColor(c):
    r=int(c[0:2],16)
    g=int(c[2:4],16)
    b=int(c[4:6],16)
    return QColor(r,g,b)


class ViewerLexer(Python3Lexer):
    extra_vars = {'vi', 'all_modules', 'image_utils', 'viewer'}
    extra_callables = {'update_workEnv', 'clear_workEnv', 'get_module', 'get_batch_manager', 'get_workEnv', 'get_image, get_meta'}
    extra_classes = {'ViewerWorkEnv', 'ImgData'}

    def get_tokens_unprocessed(self, text):
        for index, token, value in Python3Lexer.get_tokens_unprocessed(self, text):
            if token is Name and value in self.extra_vars:
                yield index, Name.Namespace, value
            elif token is Name and value in self.extra_callables:
                yield index, Name.Function, value
            elif token is Name and value in self.extra_classes:
                yield index, Name.Class, value
            else:
                yield index, token, value

class QFormatter(Formatter):
    def __init__(self):
        Formatter.__init__(self)
        self.data = []

        self.styles = {}
        for token, style in self.style:
            qtf = QTextCharFormat()

            if style['color']:
                qtf.setForeground(hex2QColor(style['color']))
            if style['bgcolor']:
                qtf.setBackground(hex2QColor(style['bgcolor']))
            if style['bold']:
                qtf.setFontWeight(QFont.Bold)
            if style['italic']:
                qtf.setFontItalic(True)
            if style['underline']:
                qtf.setFontUnderline(True)
            self.styles[str(token)] = qtf

    def format(self, tokensource, outfile):
        global styles
        self.data = []


        for ttype, value in tokensource:
            l = len(value)
            t = str(ttype)
            self.data.extend([self.styles[t], ] * l)


class Highlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super(Highlighter, self).__init__(parent)
        self.formatter = QFormatter()
        self.lexer = ViewerLexer()

    def highlightBlock(self, text):

        cb = self.currentBlock()
        p = cb.position()

        text = self.document().toPlainText()

        highlight(text, self.lexer, self.formatter)

        for i in range(len(text)):
            try:
                self.setFormat(i, 1, self.formatter.data[p + i])
            except IndexError:
                pass



from PyQt5 import QtCore, QtGui, QtWidgets
from glob import glob
import os
from matplotlib import cm as matplotlib_color_map
from typing import *
from ..pyqtgraphCore.functions import mkColor
import numpy as np
from collections import OrderedDict
from warnings import warn

qual_cmaps = ['Pastel1', 'Pastel2', 'Paired', 'Accent', 'Dark2', 'Set1',
              'Set2', 'Set3', 'tab10', 'tab20', 'tab20b', 'tab20c']


def auto_colormap(
        n_colors: int,
        cmap: str = 'hsv',
        output: str = 'mpl',
        spacing: str = 'uniform',
        alpha: float = 1.0
    ) \
        -> List[Union[QtGui.QColor, np.ndarray, str]]:
    """
    If non-qualitative map: returns list of colors evenly spread through the chosen colormap.
    If qualitative map: returns subsequent colors from the chosen colormap

    :param n_colors: Numbers of colors to return
    :param cmap:     name of colormap

    :param output:   option: 'mpl' returns RGBA values between 0-1 which matplotlib likes,
                     option: 'pyqt' returns QtGui.QColor instances that correspond to the RGBA values
                     option: 'bokeh' returns hex strings that correspond to the RGBA values which bokeh likes

    :param spacing:  option: 'uniform' returns evenly spaced colors across the entire cmap range
                     option: 'subsequent' returns subsequent colors from the cmap

    :param alpha:    alpha level, 0.0 - 1.0

    :return:         List of colors as either ``QColor``, ``numpy.ndarray``, or hex ``str`` with length ``n_colors``
    """

    valid = ['mpl', 'pyqt', 'bokeh']
    if output not in valid:
        raise ValueError(f'output must be one {valid}')

    valid = ['uniform', 'subsequent']
    if spacing not in valid:
        raise ValueError(f'spacing must be one of either {valid}')

    if alpha < 0.0 or alpha > 1.0:
        raise ValueError('alpha must be within 0.0 and 1.0')

    cm = matplotlib_color_map.get_cmap(cmap)
    cm._init()

    if output == 'pyqt':
        lut = (cm._lut * 255).view(np.ndarray)
    else:
        lut = (cm._lut).view(np.ndarray)

    lut[:, 3] *= alpha

    if spacing == 'uniform':
        if not cmap in qual_cmaps:
            cm_ixs = np.linspace(0, 210, n_colors, dtype=int)
        else:
            if n_colors > len(lut):
                raise ValueError('Too many colors requested for the chosen cmap')
            cm_ixs = np.arange(0, len(lut), dtype=int)
    else:
        cm_ixs = range(n_colors)

    colors = []
    for ix in range(n_colors):
        c = lut[cm_ixs[ix]]

        if output == 'pyqt':
            colors.append(mkColor(c))

        elif output == 'bokeh':
            c = tuple(c[:3] * 255)
            hc = '#%02x%02x%02x' % tuple(map(int, c))
            colors.append(hc)

        else:  # mpl
            colors.append(c)

    return colors


def get_colormap(labels: iter, cmap: str, **kwargs) -> OrderedDict:
    """
    Get a dict for mapping labels onto colors

    Any kwargs are passed to auto_colormap()

    :param labels:  labels for creating a colormap. Order is maintained if it is a list of unique elements.
    :param cmap:    name of colormap

    :return:        dict of labels as keys and colors as values
    """
    if not len(set(labels)) == len(labels):
        labels = list(set(labels))
    else:
        labels = list(labels)

    colors = auto_colormap(len(labels), cmap, **kwargs)

    return OrderedDict(zip(labels, colors))


def map_labels_to_colors(labels: iter, cmap: str, **kwargs) -> list:
    """
    Map labels onto colors according to chosen colormap

    Any kwargs are passed to auto_colormap()

    :param labels:  labels for mapping onto a colormap
    :param cmap:    name of colormap
    :return:        list of colors mapped onto the labels
    """
    mapper = get_colormap(labels, cmap, **kwargs)
    return list(map(mapper.get, labels))


class ColormapListWidget(QtWidgets.QListWidget):
    signal_colormap_changed = QtCore.pyqtSignal(str)

    def __init__(self, parent):
        QtWidgets.QListWidget.__init__(self, parent=parent)
        self.populate_colormaps()
        self.currentItemChanged.connect(self.emit_colormap_changed)
        self.setToolTip('Colormap picker')
        self.setCurrentRow(0)

    @property
    def current_cmap(self) -> str:
        return self.currentItem().text()

    def set_cmap(self, cmap: str):
        item = self.findItems(cmap, QtCore.Qt.MatchExactly)[0]
        ix = self.indexFromItem(item)
        self.setCurrentIndex(ix)

    def get_cmap(self) -> str:
        return self.currentItem().text()

    def emit_colormap_changed(self, item: QtWidgets.QListWidgetItem):
        self.signal_colormap_changed.emit(item.text())

    def populate_colormaps(self):
        mpath = os.path.abspath(__file__)
        mdir = os.path.dirname(mpath)
        cmap_img_files = glob(mdir + '/colormaps/*.png')
        cmap_img_files.sort()

        for path in cmap_img_files:
            img = QtGui.QIcon(path)
            item = QtWidgets.QListWidgetItem(os.path.basename(path)[:-4])
            item.setIcon(img)
            self.addItem(item)
        self.setIconSize(QtCore.QSize(120, 25))


class WidgetEntry:
    def __init__(self, setter: callable, getter: callable, name: str):
        self.setter = setter
        self.getter = getter
        self.name = name


class WidgetRegistry:
    """
    Register widgets to conveniently store and restore their states
    """

    def __init__(self):
        self.widgets = dict()

    def register(self, widget: QtWidgets.QWidget, setter: callable, getter: callable, name: str):
        """
        Register a widget. The `setter` and `getter` methods must be interoperable

        :param widget: widget to register
        :type widget: QtWidgets.QWidget

        :param setter: widget's method to use for setting its value
        :type setter: callable

        :param getter: widget's method to use for getting its value. This value must be settable through the specified "setter" method
        :type getter: callable

        :param name: a name for this widget
        :type name: str
        """
        if (not callable(setter)) or (not callable(getter)):
            raise TypeError('setter and getter must be callable')

        self.widgets[widget] = WidgetEntry(setter=setter, getter=getter, name=name)

    def get_state(self) -> dict:
        """Get a dict of states for all registered widgets"""
        s = dict()

        for w in self.widgets.keys():
            name = self.widgets[w].name
            s[name] = self.widgets[w].getter()

        return s

    def set_state(self, state: dict):
        """Set all registered widgets from a dict"""
        for w in self.widgets.keys():
            name = self.widgets[w].name

            # account for using old saved state files whilst control widgets change
            if name not in state.keys():
                warn(f'State not available for widget: {name}\n{w}')
                continue

            s = state[name]

            self.widgets[w].setter(s)
